﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: HatchingFilter.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib;
using VideoLib.Frames;
using VideoLib.OpenGL;

namespace RGBZcamera.Library.Hatching
{
	internal struct Sprite
	{
		public Sprite(float x, float y)
		{
			position = new Vector2(x, y);
		}

		public Vector2 position;

		public float X { get { return position.X; } set { position.X = value; } }
		public float Y { get { return position.Y; } set { position.Y = value; } }

		public override string ToString()
		{
			return string.Format("Sprite({0}, {1})", X, Y);
		}
	}

	public class HatchingFilter : Node, IOpenGLNode
	{
		public HatchingFilter()
		{
			TargetCoverage    = 1;
			CoverageSize      = 4;
			SpriteSize        = 10;
			AddThreshold      = 1;
			DeleteThreshold   = 3;
			BackgroundPlane   = new Vector4(0, 0, -1, 1); // default: plane at 1m depth from camera
			PlaneThreshold    = float.NegativeInfinity; // disable thresholding
			BlackOrWhite      = true;
			Attenuation       = 0;// 1.1f;
			LightDirection    = new Vector3(0, 0, -1);
			JitterPosition    = 0;
			JitterRotation    = 0;
			OrientByColour    = false;
			SmoothOrientation = false;

			sprites   = new List<Sprite>();
			random    = new Random(0x12345678);
			depthBlur = new DepthBlurNode() { Topology = null };
			depthBilateral = new DepthBilateralNode() { Topology = null };

			AddInputPins("colour", "points", "flow");
			AddOutputPin("out");
		}

		[DefaultValue(1.0f), Description("Desired number of overlapping sprites at each pixel."), DisplayName("Target coverage"), Category("Sprite distribution")]
		public float TargetCoverage { get; set; }

		[DefaultValue(4.0f), Description("Size of sprites for the purpose of coverage computation."), DisplayName("Coverage size"), Category("Sprite distribution")]
		public float CoverageSize { get; set; }

		[DefaultValue(10.0f), Description("Size of sprites."), DisplayName("Sprite size"), Category("Sprite distribution")]
		public float SpriteSize { get; set; }

		[DefaultValue(1.0f), Description("Add sprites if the coverage falls below this threshold (1 means target coverage)."), DisplayName("Add threshold"), Category("Sprite distribution")]
		public float AddThreshold { get; set; }

		[DefaultValue(3.0f), Description("Delete sprites if the coverage exceeds this threshold (1 means target coverage)."), DisplayName("Delete threshold"), Category("Sprite distribution")]
		public float DeleteThreshold { get; set; }

		[Description("Background plane equation for background subtraction (Xx+Yy+Zy+W=0)."), DisplayName("Background plane"), Category("Rendering")]
		public Vector4 BackgroundPlane { get; set; }

		[DefaultValue(float.NegativeInfinity), Description("Threshold for background segmentation."), DisplayName("Plane threshold"), Category("Rendering")]
		public float PlaneThreshold { get; set; }

		[DefaultValue(true), Description("Selects background colour and blending mode (false=black, true=white)."), DisplayName("Black or white?"), Category("Rendering")]
		public bool BlackOrWhite { get; set; }

		[DefaultValue(0.0f), Description("Attenuates strokes with normals close to the light direction."), DisplayName("Stroke attenuation"), Category("Rendering")]
		public float Attenuation { get; set; }

		[Description("Light direction used for stroke attenuation."), DisplayName("Light direction"), Category("Rendering")]
		public Vector3 LightDirection { get; set; }

		[DefaultValue(0.0f), Description("Maximum random jitter applied to each sprite's 2D position."), DisplayName("Jitter position"), Category("Rendering")]
		public float JitterPosition { get; set; }

		[DefaultValue(0.0f), Description("Maximum random jitter applied to each sprite's orientation (in degrees)."), DisplayName("Jitter rotation"), Category("Rendering")]
		public float JitterRotation { get; set; }

		[DefaultValue(false), Description("Orient sprites using intensity instead of depth map."), DisplayName("Orient by colour?"), Category("Orientation")]
		public bool OrientByColour { get; set; }

		[DefaultValue(false), Description("Smooth sprite orientations by blurring the depth map (or image)."), DisplayName("Smooth orientation"), Category("Orientation")]
		public bool SmoothOrientation { get; set; }

		public bool Bilateral { get; set; }

		public bool ShouldSerializeBackgroundPlane() { return !BackgroundPlane.Equals(new Vector4(0, 0, -1, 1)); }
		public bool ShouldSerializeLightDirection() { return !LightDirection.Equals(new Vector3(0, 0, -1)); }
		public void ResetBackgroundPlane() { BackgroundPlane = new Vector4(0, 0, -1, 1); }
		public void ResetLightDirection() { LightDirection = new Vector3(0, 0, -1); }

		private List<Sprite> sprites;
		private Random random;
		private DepthBlurNode depthBlur;
		private DepthBilateralNode depthBilateral;

		private int coverageFBO, coverageIBO, coverageVBO;
		private int advectionFBO, advectionProgram;
		private int renderQuadsFBO, renderQuadsProgram;

		public unsafe override void Init()
		{
			// generate FBOs
			GL.GenFramebuffers(1, out coverageFBO);
			GL.GenFramebuffers(1, out advectionFBO);
			GL.GenFramebuffers(1, out renderQuadsFBO);

			// index buffer object: just 0, 1, ..., 2^16-1
			GL.GenBuffers(1, out coverageIBO);
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, coverageIBO);
			var indices = Enumerable.Range(0, 2 << 16).Select(x => (ushort)x).ToArray();
			GL.BufferData(BufferTarget.ElementArrayBuffer, (IntPtr)((2 << 16) * sizeof(ushort)), indices, BufferUsageHint.StaticDraw);
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, 0);

			// vertex buffer object: reserve space for 2^16 sprites, which are updated every frame
			GL.GenBuffers(1, out coverageVBO);
			GL.BindBuffer(BufferTarget.ArrayBuffer, coverageVBO);
			GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)((2 << 16) * Vector3.SizeInBytes), IntPtr.Zero, BufferUsageHint.StreamDraw);
			GL.BindBuffer(BufferTarget.ArrayBuffer, 0);

			if (!GL.GetString(StringName.Extensions).Contains("EXT_geometry_shader4"))
				throw new NotSupportedException("Your video card does not support EXT_geometry_shader4. Please update your drivers.");

			// compile and link programs
			advectionProgram = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Hatching.Shaders.SpriteAdvection.frag")
			);

			renderQuadsProgram = GL.CreateProgram();
			GL.Ext.ProgramParameter(renderQuadsProgram, ExtGeometryShader4.GeometryInputTypeExt,  (int)All.Points);
			GL.Ext.ProgramParameter(renderQuadsProgram, ExtGeometryShader4.GeometryOutputTypeExt, (int)All.TriangleStrip);
			GL.Ext.ProgramParameter(renderQuadsProgram, ExtGeometryShader4.GeometryVerticesOutExt, 4);
			Utils.LinkShaderProgram(renderQuadsProgram,
				Utils.CreateShaderFromResource("Shaders.Mesh.vert"),
				Utils.CreateShaderFromResource("Hatching.Shaders.RenderQuads.vert"),
				Utils.CreateShaderFromResource("Hatching.Shaders.RenderQuads.geom"),
				Utils.CreateShaderFromResource("Hatching.Shaders.RenderQuads.frag")
			);

			// set texture sampler uniforms
			GL.UseProgram(advectionProgram);
			GL.Uniform1(GL.GetUniformLocation(advectionProgram, "sprites"), 0);
			GL.Uniform1(GL.GetUniformLocation(advectionProgram, "flow"),    1);

			GL.UseProgram(renderQuadsProgram);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "colour"), 0);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "points"), 1);
			GL.UseProgram(0);

			depthBlur.Init();
			depthBilateral.Init();

			base.Init();
		}

		public override void Process()
		{
			var sw = Stopwatch.StartNew();
			var tt = new Dictionary<string, long>();

			// get inputs
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);
			AssertFrameProperties(textures, FrameProps.Size);

			// advect sprites
			AdvectSprites(textures["flow"]);
			if (Verbose) { tt.Add("adv", sw.ElapsedMilliseconds); sw = Stopwatch.StartNew(); }

			// determine coverage of sprites
			var coverageTex = DrawSpriteCoverage(textures["points"].Width, textures["points"].Height);
			if (Verbose) { tt.Add("cov", sw.ElapsedMilliseconds); sw = Stopwatch.StartNew(); }

			var coverage = coverageTex.Texture.ReadFrame(VideoLib.Frames.PixelFormat.Float);
			coverageTex.RefDec(); coverageTex = null;
			if (Verbose) { tt.Add("ro", sw.ElapsedMilliseconds); sw = Stopwatch.StartNew(); }

			// update sprites: insert & delete
			UpdateSprites(coverage);
			if (Verbose) { tt.Add("upd", sw.ElapsedMilliseconds); sw = Stopwatch.StartNew(); }

			//// visualise sprite positions (comment out DrawSpriteQuads below)
			//var covSize = CoverageSize;
			//CoverageSize = 2;
			//var rendered = DrawSpriteCoverage(textures["points"].Width, textures["points"].Height);
			//rendered.SequenceNumber = textures["colour"].SequenceNumber;
			//CoverageSize = covSize;

			// render sprites
			var rendered = DrawSpriteQuads(textures);
			if (Verbose) { tt.Add("rnd", sw.ElapsedMilliseconds); sw = Stopwatch.StartNew(); }
			
			// output the result and clean up
			PushFrame(rendered);
			rendered.RefDec();
			textures.RefDec();

			if (Verbose)
			{
				// print timing data
				Console.WriteLine("Timing: {0} | total:{1:D3} ms",
					string.Join("  ", tt.Select(x => string.Format("{0}:{1:D3}", x.Key, x.Value)).ToArray()),
					tt.Select(x => x.Value).Sum()
				);
			}
		}

		private unsafe void AdvectSprites(TextureFrame flow)
		{
			// skip if there are no sprites
			if (sprites.Count == 0)
				return;

			// create sprite texture (need to tile in 2D)
			int width  = Math.Min(1024, sprites.Count);
			int height = (sprites.Count + width - 1) / width;
			var inputSprites = new Texture(width, height)
			{
				InternalFormat = PixelInternalFormat.Rgba32f,
				Filter = TextureMagFilter.Nearest,
				Wrap = TextureWrapMode.ClampToEdge
			};

			// fill sprite texture with sprite positions in previous frame
			Vector4[] values = new Vector4[width * height];
			for (int i = 0; i < sprites.Count; i++) values[i].Xy = sprites[i].position;
			inputSprites.Create(values, width * Vector4.SizeInBytes, OpenTK.Graphics.OpenGL.PixelFormat.Rgba, PixelType.Float);

			// prepare texture for new sprite positions
			var outputSprites = new Texture(inputSprites.Width, inputSprites.Height) { InternalFormat = PixelInternalFormat.Rgba32f };
			outputSprites.Alloc();

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, advectionFBO);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, outputSprites, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			GL.Disable(EnableCap.DepthTest);
			GL.UseProgram(advectionProgram);
			flow.Bind(TextureUnit.Texture1);
			inputSprites.Bind(TextureUnit.Texture0);

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();

			GL.Viewport(0, 0, inputSprites.Width, inputSprites.Height);

			// draw screen filling quad with appropriate texture coordinates
			GL.Begin(BeginMode.Quads);
			GL.TexCoord2(0, 0); GL.Vertex2(-1, -1);
			GL.TexCoord2(1, 0); GL.Vertex2( 1, -1);
			GL.TexCoord2(1, 1); GL.Vertex2( 1,  1);
			GL.TexCoord2(0, 1); GL.Vertex2(-1,  1);
			GL.End();

			GL.UseProgram(0);
			GL.Enable(EnableCap.DepthTest);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
			flow.Unbind();
			inputSprites.Unbind();
			inputSprites.Free();

			// read out new sprite positions from texture
			var newSprites = outputSprites.ReadFrame(VideoLib.Frames.PixelFormat.FloatRGBA);
			outputSprites.Free();

			// update sprite positions
			fixed (byte* pByte = newSprites.Data)
			{
				Vector4* pNewSprites = (Vector4*)pByte;

				for (int i = 0; i < sprites.Count; i++)
				{
					var sprite = sprites[i];
					sprite.position = pNewSprites[i].Xy;
					sprites[i] = sprite;
				}
			}
		}

		private unsafe TextureFrame DrawSpriteCoverage(int width, int height)
		{
			// allocate the output texture
			var coverage = new TextureFrame(width, height);
			coverage.Alloc();

			if (sprites.Count == 0)
				return coverage;

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, coverageFBO);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, coverage.Texture, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			GL.Disable(EnableCap.DepthTest);
			GL.Disable(EnableCap.Lighting);

			// use additive blend (with alpha)
			GL.Enable(EnableCap.Blend);
			GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.One);

			GL.MatrixMode(MatrixMode.Projection);
			var projMat = Matrix4.CreateOrthographicOffCenter(0, coverage.Width, 0, coverage.Height, 1, 10);
			GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();

			GL.Viewport(0, 0, coverage.Width, coverage.Height);

			GL.ClearColor(Color.Black);
			GL.Clear(ClearBufferMask.ColorBufferBit);

			//// draw point using immediate mode (kind of slow: 60-70 ms for 25K)
			//GL.PointSize(pointSize);
			//GL.Begin(BeginMode.Points);
			//foreach (var sprite in sprites)
			//{
			//    GL.Color4(1.0, 1.0, 1.0, 1.0);
			//    GL.Vertex3(sprite.X, sprite.Y, -2);
			//}
			//GL.End();

			GL.BindBuffer(BufferTarget.ArrayBuffer, coverageVBO);
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, coverageIBO);

			// update VBO with current sprites
			var vertexBuffer = GL.MapBuffer(BufferTarget.ArrayBuffer, BufferAccess.WriteOnly);
			Vector3* spriteVBO = (Vector3*)vertexBuffer.ToPointer();

			int index = 0;
			foreach (var sprite in sprites)
				spriteVBO[index++] = new Vector3(sprite.X, sprite.Y, -2.0f);

			GL.UnmapBuffer(BufferTarget.ArrayBuffer);

			// sets the vertex pointer to the currently bound array describing how the data is stored
			GL.PushClientAttrib(ClientAttribMask.ClientVertexArrayBit);
			GL.VertexPointer(3, VertexPointerType.Float, Vector3.SizeInBytes, IntPtr.Zero);
			GL.EnableClientState(ArrayCap.VertexArray);

			// draw points from VBO
			GL.PointSize(CoverageSize);
			GL.DrawElements(BeginMode.Points, sprites.Count, DrawElementsType.UnsignedShort, IntPtr.Zero);

			// VBO cleanup
			GL.PopClientAttrib();
			GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, 0);

			// general clean up
			GL.Disable(EnableCap.Blend);
			GL.Enable(EnableCap.DepthTest);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);

			return coverage;
		}

		private unsafe void UpdateSprites(BitmapFrame coverage)
		{
			// TODO: fade sprites in/out
			fixed (byte* pByte = coverage.Data)
			{
				float* pCoverage = (float*)pByte;

				// adding new sprites
				for (int y = 0; y < coverage.Height; y++)
				{
					for (int x = 0; x < coverage.Width; x++)
					{
						var currentCoverage = pCoverage[y * coverage.Width + x];
						if (currentCoverage < AddThreshold * TargetCoverage)
						{
							// probabilistically add sprite
							if (random.NextDouble() < (TargetCoverage - currentCoverage) / (CoverageSize * CoverageSize) && sprites.Count < (2 << 16))
								sprites.Add(new Sprite(x, y));
						}
					}
				}

				// removing sprites
				for (int i = sprites.Count - 1; i >= 0; i--)
				{
					var sprite = sprites[i];

					// always remove sprites that fall outside the image
					if (sprite.X < 0 || sprite.X >= coverage.Width || sprite.Y < 0 || sprite.Y >= coverage.Height)
					{
						sprites.RemoveAt(i);
						continue;
					}

					// check for areas that are too dense
					var currentCoverage = pCoverage[(int)sprite.Y * coverage.Width + (int)sprite.X];
					if (currentCoverage > DeleteThreshold * TargetCoverage)
					{
						// probabilistically delete sprite
						if(random.NextDouble() > TargetCoverage / currentCoverage)
							sprites.RemoveAt(i);
					}
				}
			}
		}

		private TextureFrame DrawSpriteQuads(Dictionary<string, TextureFrame> inputs)
		{
			TextureFrame points = inputs[OrientByColour ? "colour" : "points"];
			points.RefInc();
			if (SmoothOrientation)
			{
			    // filter points to blur depth
				depthBlur.Sigma = SpriteSize / 4;
				depthBilateral.SpatialSigma = SpriteSize / 4;
				TextureFrame smoothedPoints = null;
				if(!Bilateral) smoothedPoints = depthBlur.Draw(new Dictionary<string, TextureFrame> { { "in", points } });
				else smoothedPoints = depthBilateral.Draw(new Dictionary<string, TextureFrame> { { "in", points } });
				points.RefDec();
				points = smoothedPoints;
			}
			
			// allocate the output texture
			var result = new TextureFrame(inputs["colour"].Width, inputs["colour"].Height) { SequenceNumber = inputs["colour"].SequenceNumber };
			result.Alloc();

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, renderQuadsFBO);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, result.Texture, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			GL.Disable(EnableCap.DepthTest);
			GL.Disable(EnableCap.Lighting);

			// use additive blend (with alpha)
			GL.Enable(EnableCap.Blend);
			GL.UseProgram(renderQuadsProgram);
			if(BlackOrWhite) GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);
			else             GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.One);

			// update uniforms
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "pointSize"),         SpriteSize);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "targetCoverage"),    TargetCoverage);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "intensityAdjust"),   (float)(SpriteSize * SpriteSize) / (CoverageSize * CoverageSize));
			GL.Uniform4(GL.GetUniformLocation(renderQuadsProgram, "backgroundPlane"),   BackgroundPlane);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "planeThreshold"),    PlaneThreshold);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "blackOrWhite"),      BlackOrWhite ? 1 : 0);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "attenuationFactor"), Attenuation);
			GL.Uniform3(GL.GetUniformLocation(renderQuadsProgram, "lightDirection"),    LightDirection);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "jitterPosition"),    JitterPosition);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "jitterRotation"),    JitterRotation);
			GL.Uniform1(GL.GetUniformLocation(renderQuadsProgram, "input"),             OrientByColour ? 4 : (SmoothOrientation ? 3 : 2));

			// bind textures
			//inputs["points"].Bind(TextureUnit.Texture1); // unblurred depth map
			points.Bind(TextureUnit.Texture1);
			inputs["colour"].Bind(TextureUnit.Texture0);

			GL.MatrixMode(MatrixMode.Projection);
			var projMat = Matrix4.CreateOrthographicOffCenter(0, result.Width, 0, result.Height, 10, -10);
			GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();

			GL.Viewport(0, 0, result.Width, result.Height);

			GL.ClearColor(BlackOrWhite ? Color.White : Color.Black);
			GL.Clear(ClearBufferMask.ColorBufferBit);

			// draw screen filling quad with appropriate texture coordinates
			//GL.PointSize(10);
			GL.Begin(BeginMode.Points);
			foreach (var sprite in sprites)
			{
				//GL.Color4(1.0, 1.0, 1.0, 1.0);
				GL.Vertex3(sprite.X, sprite.Y, 0);
			}
			GL.End();

			// clean up
			GL.UseProgram(0);
			//inputs["points"].Unbind(); // unblurred
			points.Unbind();
			inputs["colour"].Unbind();
			GL.Enable(EnableCap.DepthTest);
			GL.Enable(EnableCap.Lighting);
			GL.Disable(EnableCap.Blend);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
			points.RefDec();

			return result;
		}
	}
}
