﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: OffExportSink.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.IO;
using System.Text;
using VideoLib;
using VideoLib.Frames;

namespace RGBZcamera.Library
{
	public class OffExportSink : Node
	{
		/// <summary>Filename template for use with String.Format, e.g. "{0:D4}.off" for "NNNN.off".</summary>
		public string Filenames { get; protected set; }

		/// <summary>Gets or sets a value indicating whether invalid triangles will be skipped.</summary>
		public bool SkipInvalidTriangles { get; set; }

		/// <summary>Initialises a new instance of the <see cref="OffExportSink"/> class.</summary>
		/// <param name="filenames">Template for filenames for use with String.Format, e.g. "{0:D4}.off" for "NNNN.off".</param>
		public OffExportSink(string filenames)
		{
			Filenames = filenames;
			SkipInvalidTriangles = true;

			AddInputPin("in");
		}

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		public override void Init()
		{
			// make sure directory exists
			MakeDirs(String.Format(Filenames, 0));
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get latest frame
			var frame = GetInputFrame();
			if (PassOnEndOfStreamFrame(frame)) return;
			var points = CastFrameTo<BitmapFrame>(frame);

			// save frame
			SaveFrame(points);
		}

		/// <summary>Makes sure that the output directory exists.</summary>
		/// <param name="filename">One image path.</param>
		private void MakeDirs(string filename)
		{
			FileInfo fileInfo = new FileInfo(filename);
			DirectoryInfo parentDir = fileInfo.Directory;
			if (!parentDir.Exists)
				parentDir.Create();
		}
		
		/// <summary>Saves the frame as a mesh using the Object File Format (OFF).</summary>
		/// <param name="frame">The frame.</param>
		public unsafe void SaveFrame(BitmapFrame frame)
		{
			string filename = String.Format(Filenames, frame.SequenceNumber);
			SaveFrame(frame, filename);
		}

		/// <summary>Saves the frame as a mesh using the Object File Format (OFF).</summary>
		/// <param name="frame">The frame.</param>
		/// <param name="filename">The filename.</param>
		public unsafe void SaveFrame(BitmapFrame frame, string filename)
		{
			AssertPixelFormat(frame, VideoLib.Frames.PixelFormat.FloatRGBA);

			// write vertices and faces to buffer first, to count them
			var sb = new StringBuilder();
			var numVertices = 0;
			var numFaces    = 0;
			fixed (byte* bytes = frame.Data)
			{
				float* points = (float*)bytes;

				// write out all vertices (also invalid ones, as this simplifies indexing of the faces)
				for (int y = 0; y < frame.Height; y++)
				{
					for (int x = 0; x < frame.Width; x++)
					{
						// flip x and y to convert from SR4000 to OpenGL coordinate system
						sb.AppendFormat("{0} {1} {2}{3}",
							-points[4 * (y * frame.Width + x) + 0],
							 points[4 * (y * frame.Width + x) + 1],
							-points[4 * (y * frame.Width + x) + 2],
							Environment.NewLine
						);
						numVertices++;
					}
				}

				// write all faces
				for (int y = 0; y < frame.Height - 1; y++)
				{
					for (int x = 0; x < frame.Width - 1; x++)
					{
						// vertex indices
						int TL = (y + 0) * frame.Width + (x + 0); // TL--TR   top-left/right
						int TR = (y + 0) * frame.Width + (x + 1); // | \  |
						int BL = (y + 1) * frame.Width + (x + 0); // |  \ |
						int BR = (y + 1) * frame.Width + (x + 1); // BL--BR   bottom-left/right

						bool TLvalid = points[4 * TL + 3] != 0.0 && (points[4 * TL + 0] != 0.0 || points[4 * TL + 1] != 0.0 || points[4 * TL + 2] != 0.0 || points[4 * TL + 3] != 1.0);
						bool TRvalid = points[4 * TR + 3] != 0.0 && (points[4 * TR + 0] != 0.0 || points[4 * TR + 1] != 0.0 || points[4 * TR + 2] != 0.0 || points[4 * TR + 3] != 1.0);
						bool BLvalid = points[4 * BL + 3] != 0.0 && (points[4 * BL + 0] != 0.0 || points[4 * BL + 1] != 0.0 || points[4 * BL + 2] != 0.0 || points[4 * BL + 3] != 1.0);
						bool BRvalid = points[4 * BR + 3] != 0.0 && (points[4 * BR + 0] != 0.0 || points[4 * BR + 1] != 0.0 || points[4 * BR + 2] != 0.0 || points[4 * BR + 3] != 1.0);

						// right and left are swapped by the x-flip above

						// first triangle (top-right)
						if (!SkipInvalidTriangles || (TRvalid && TLvalid && BLvalid))
						{
							//writer.WriteLine("3 {0} {1} {2}", TL, TR, BR);
							sb.AppendFormat("3 {0} {1} {2}{3}", TR, TL, BL, Environment.NewLine);
							numFaces++;
						}

						// second triangle (bottom-left)
						if (!SkipInvalidTriangles || (TRvalid && BLvalid && BRvalid))
						{
							//writer.WriteLine("3 {0} {1} {2}", TL, BR, BL);
							sb.AppendFormat("3 {0} {1} {2}{3}", TR, BL, BR, Environment.NewLine);
							numFaces++;
						}
					}
				}
			}

			using (var writer = new StreamWriter(filename))
			{
				// first line of the file header
				writer.WriteLine("OFF");

				// NVertices NFaces NEdges (NEdges not used or checked)
				writer.WriteLine("{0} {1} 0", numVertices, numFaces);

				// write vertices and faces
				writer.Write(sb.ToString());
			}
		}
	}
}
