﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: StereoWarper.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using RGBZcamera.Library.Config;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib;
using VideoLib.OpenGL;
using VideoLib.Vision;

namespace RGBZcamera.Library
{
	/// <summary>Implements a OpenGL-based stereo image warper.</summary>
	public partial class StereoWarper : Node, IOpenGLNode
	{
		/// <summary>Gets or sets the extrinsic parameters of the source camera (i.e. ToF).</summary>
		public ExtrinsicParameters SourceExtrinsics { get; set; }

		/// <summary>Gets or sets the intrinsic parameters of the source camera (i.e. ToF).</summary>
		public IntrinsicParameters SourceIntrinsics { get; set; }

		/// <summary>Gets or sets the extrinsic parameters of the target camera (i.e. video camera).</summary>
		public ExtrinsicParameters TargetExtrinsics { get; set; }

		/// <summary>Gets or sets the target intrinsic parameters of the target camera (i.e. video camera).</summary>
		public IntrinsicParameters TargetIntrinsics { get; set; }

		/// <summary>Gets or sets the type of the sensor.</summary>
		public SensorType SensorType { get; set; }

		/// <summary>Gets or sets the inter-ocular distance, in metres.</summary>
		[DefaultValue(0.065f), Category("Stereo settings"), Description("The distance between the two virtual cameras (in metres).")]
		public float InterOcularDistance { get; set; }

		[DefaultValue(1.0f), Category("Stereo settings"), Description("The distance to the screen plane or plane of zero disparity (in metres).")]
		public float ScreenDistance { get; set; }

		/// <summary>Gets or sets a value indicating whether the view is cropped. Default: true.</summary>
		public bool CropView { get; set; }

		/// <summary>Gets or sets the near plane used in the cropping frame computation. Default: 0.5 (m).</summary>
		public float CropNear { get; set; }

		/// <summary>Gets or sets the farplane used in the cropping frame computation. Default: 5.0 (m).</summary>
		public float CropFar { get; set; }

		/// <summary>Gets or sets a value indicating the base output resolution (true=video; false=depth). Default: true (video resolution).</summary>
		public bool OutputVideoResolution { get; set; }

		/// <summary>Get or sets how far into the background a texture sample is taken from to fill disocclusions.</summary>
		[Uniform, DefaultValue(1.0f), Category("Stereo setting"), Description("Offset for reading out background texture in disocclusion regions.")]
		public float Shift { get; set; }


		/// <summary>Constructs an OpenGL-based mesh reprojection node.</summary>
		public StereoWarper()
		{
			SensorType            = SensorType.FleaSR;
			InterOcularDistance   = 0.065f;
			ScreenDistance        = 1.0f;
			Shift                 = 1;

			// these settings need to be identical to the ReprojectionNode settings, for the stereo warping to work correctly
			OutputVideoResolution = true;
			CropView              = true;
			CropNear              = 0.5f;
			CropFar               = 5.0f;

			AddInputPins( "pointsIn",  "colourIn");
			AddOutputPins("pointsOutL", "pointsOutR", "colourOutL", "colourOutR");
		}

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			GL.GenBuffers(1, out VertexBufferID);
			GL.GenBuffers(1, out IndexBufferID);

			// load, compile and link shaders
			MeshProgramID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("Shaders.Mesh.vert"),
				Utils.CreateShaderFromResource("Shaders.StereoWarper.vert"),
				Utils.CreateShaderFromResource("Shaders.StereoWarper.frag")
			);

			// set uniforms
			GL.UseProgram(MeshProgramID);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "coordinates"), 0); // texture unit 0
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "video"), 1); // texture unit 1
			GL.UseProgram(0);
		}

		public override void Process()
		{
			// get latest frames
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textureFrames = CastFramesTo<TextureFrame>(frames);

			PointCloudTexture = textureFrames["pointsIn"];
			VideoTexture      = textureFrames["colourIn"];

			// determine output resolution
			if(OutputVideoResolution) VideoSize = new Size(VideoTexture.Width, VideoTexture.Height);
			else                      VideoSize = new Size(PointCloudTexture.Width, PointCloudTexture.Height);

			// update mesh
			MeshTools.UpdateBaseMeshVertexBuffer(PointCloudTexture.Width, PointCloudTexture.Height, VertexBufferID, out NumVertices);
			MeshTools.UpdateBaseMeshIndexBuffer( PointCloudTexture.Width, PointCloudTexture.Height, IndexBufferID,  out NumIndices);

			// convert from linear to nearest texture filtering (is that necessary?)
			PointCloudTexture.Filter = TextureMagFilter.Nearest;
			PointCloudTexture.ApplyParameters();

			ComputeCropFrame(PointCloudTexture);
			SetupFramebufferObjects();

			var left = DrawMesh(textureFrames["colourIn"].SequenceNumber, textureFrames["pointsIn"].SequenceNumber, InterOcularDistance / 2);
			PushFrame("pointsOutL", left[0]);
			PushFrame("colourOutL", left[1]);
			left.RefDec();

			var right = DrawMesh(textureFrames["colourIn"].SequenceNumber, textureFrames["pointsIn"].SequenceNumber, -InterOcularDistance / 2);
			PushFrame("pointsOutR", right[0]);
			PushFrame("colourOutR", right[1]);
			right.RefDec();

			textureFrames.RefDec();
		}

		#region Cropping computations
//=================================================================================================

		/// <summary>Cropping region relative to video frame (OpenGL: x is right, y is up).</summary>
		private Rectangle Crop;
		private Size CropSize;

		private void ComputeCropFrame(Texture pointCloud)
		{
			// check if we need to recompute the crop rectangle
			if (pointCloud.Width == CropSize.Width && pointCloud.Height == CropSize.Height)
				return;

			// check that we have all parameters
			if (SourceExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of source camera not provided.");
			if (SourceIntrinsics == null) throw new InvalidOperationException("Intrinsic parameters of source camera not provided.");
			if (TargetExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of target camera not provided.");
			if (TargetIntrinsics == null) throw new InvalidOperationException("Intrinsic parameters of target camera not provided.");

			// This is how it works:
			// ---------------------
			// The cropping frame is determined by the frustum spanned by the time-of-flight sensor.
			// This is determined by back-projecting the outer corners of the depth image for both
			// the near (0.5 m) and the far planes (5 m), and then projecting them onto the target
			// camera's image plane. We then compute the largest axis-aligned rectangle that always
			// fits inside the projected mesh.

			// size of depth/distance map
			int xmin = 0; int xmax = 0; int ymin = 0; int ymax = 0;
			switch (SensorType)
			{
				case SensorType.FleaSR:
					xmax = 176 - 1; // pointCloud.Width - 1;
					ymax = 144 - 1; // pointCloud.Height - 1;
					break;

				case SensorType.Kinect:
					xmax = 640 - 1; // pointCloud.Width - 1;
					ymax = 480 - 1; // pointCloud.Height - 1;
					break;
			}

			// corner pixels of distance image, with minimum and maximum depth
			var cornerPixels = new Vector3[] {
				new Vector3(xmin, ymin, CropNear), new Vector3(xmin, ymax, CropNear), new Vector3(xmax, ymin, CropNear), new Vector3(xmax, ymax, CropNear),
				new Vector3(xmin, ymin, CropFar ), new Vector3(xmin, ymax, CropFar ), new Vector3(xmax, ymin, CropFar ), new Vector3(xmax, ymax, CropFar ),
			};

			// compute coordinates of corners in source camera reference frame
			var corners = cornerPixels.Select(x => DistanceToPoint(x, SourceIntrinsics));

			// transform to target view and project
			var invSourceView = SourceExtrinsics.ToMatrix4(); invSourceView.Invert();
			var P = TargetIntrinsics.ToMatrix4() * TargetExtrinsics.ToMatrix4() * invSourceView; P.Transpose();
			var screenPoints = corners.Select(x => Vector4.Transform(x, P));

			// normalise to get screen coordinates
			var screenCoords = screenPoints.Select(x => new Vector2(x.X / x.Z, x.Y / x.Z));

			//// find the smallest rectangle that always contains the mesh (fit outer)
			//int cxmin = screenCoords.Select(x => (int)Math.Floor(x.X)).Min();
			//int cxmax = screenCoords.Select(x => (int)Math.Ceiling(x.X)).Max();
			//int cymin = screenCoords.Select(x => (int)Math.Floor(x.Y)).Min();
			//int cymax = screenCoords.Select(x => (int)Math.Ceiling(x.Y)).Max();

			// find the largest rectangle that always fits inside the mesh (fit inner)
			int cxmin = (int)Math.Floor(  new int[] { 0, 1, 4, 5 }.Select(i => screenCoords.ElementAt(i).X).Max());
			int cxmax = (int)Math.Ceiling(new int[] { 2, 3, 6, 7 }.Select(i => screenCoords.ElementAt(i).X).Min());
			int cymin = (int)Math.Floor(  new int[] { 0, 2, 4, 6 }.Select(i => screenCoords.ElementAt(i).Y).Max());
			int cymax = (int)Math.Ceiling(new int[] { 1, 3, 5, 7 }.Select(i => screenCoords.ElementAt(i).Y).Min());

			// flip Y for OpenGL
			int tmp = cymax;
			cymax = VideoSize.Height - cymin - 1;
			cymin = VideoSize.Height - tmp - 1;

			CropSize = new Size(pointCloud.Width, pointCloud.Height);
			if(CropView) Crop = new Rectangle(cxmin, cymin, cxmax - cxmin + 1, cymax - cymin + 1);
			else         Crop = new Rectangle(0, 0, VideoSize.Width, VideoSize.Height); // no cropping
		}

		/// <summary>Back-project point (pix_x, pix_y, distance) to a homogeneous 3D point.</summary>
		/// <param name="p">The p.</param>
		/// <returns></returns>
		private Vector4 DistanceToPoint(Vector3 p, IntrinsicParameters K)
		{
			// inv([Fx 0 Cx; 0 Fy Cy; 0 0 1]) = [iFx 0 iCx; 0 iFy iCy; 0 0 1]
			float iFx = 1.0f / K.Fx;
			float iFy = 1.0f / K.Fy;
			float iCx = -K.Cx / K.Fx;
			float iCy = -K.Cy / K.Fy;

			// inv(K) * [x y 1]' = [iFx 0 iCx; 0 iFy iCy; 0 0 1] * [x y 1]'
			float x3 = iFx * p.X + iCx;
			float y3 = iFy * p.Y + iCy;
			//float z3 = 1;

			// scale point to have desired distance along ray
			// p = p / norm(p) * depth
			float scale = p.Z / (float)Math.Sqrt(x3 * x3 + y3 * y3 + 1); // z3 = 1
			return new Vector4(x3 * scale, y3 * scale, scale, 1); // z3 = 1
		}
		
//=================================================================================================
		#endregion

		#region Framebuffer Object (off-screen rendering)
//=================================================================================================

		private int FramebufferID;
		private int DepthRenderbufferID;
		private Size FramebufferSize;

		private void SetupFramebufferObjects()
		{
			// correct size already? nothing to do then ...
			if (Crop.Size == FramebufferSize) return;

			// delete any existing buffers
			if (DepthRenderbufferID != 0) GL.DeleteRenderbuffers(1, ref DepthRenderbufferID);
			if (FramebufferID       != 0) GL.DeleteFramebuffers(1, ref FramebufferID);

			// generate a new depth renderbuffer for depth
			GL.GenRenderbuffers(1, out DepthRenderbufferID);
			GL.BindRenderbuffer(RenderbufferTarget.Renderbuffer, DepthRenderbufferID);
			GL.RenderbufferStorage(RenderbufferTarget.Renderbuffer, RenderbufferStorage.DepthComponent32f, Crop.Width, Crop.Height);
			GL.BindRenderbuffer(RenderbufferTarget.Renderbuffer, 0);

			// generate a new framebuffer objects for point cloud
			GL.GenFramebuffers(1, out FramebufferID);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferRenderbuffer(FramebufferTarget.Framebuffer, FramebufferAttachment.DepthAttachment, RenderbufferTarget.Renderbuffer, DepthRenderbufferID);

			//Utils.CheckFboStatus(); // not complete, as colour buffers are missing

			FramebufferSize = Crop.Size;
		}
		
//=================================================================================================
		#endregion
		
		#region Mesh drawing methods
//=================================================================================================

		private Texture PointCloudTexture;
		private Texture VideoTexture;
		private int VertexBufferID;
		private int IndexBufferID;
		private int MeshProgramID;
		private int NumVertices;
		private int NumIndices;
		private Size VideoSize;

		private void SetupMeshMatrices(float deltaX)
		{
			//if (SourceExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of source camera not provided.");
			//if (TargetExtrinsics == null) throw new InvalidOperationException("Extrinsic parameters of target camera not provided.");
			if (TargetIntrinsics == null) throw new InvalidOperationException("Intrinsic parameters of target camera not provided.");


			// Step 1. Projection matrix from intrinsic parameters --------------------------------

			// this shift moves both stereo images horizontally relative to each other,
			// so that the plane at ScreenDistance has zero disparity, i.e. becomes the plane on the screen
			var cropShift = (int)(TargetIntrinsics.Fx * deltaX / ScreenDistance + 0.5f);

			Matrix4 projMat = Utils.CreateCalibratedProjectionMatrix(
				VideoSize.Width, VideoSize.Height,
				Crop.X + cropShift, Crop.Y, Crop.X + Crop.Width + cropShift, Crop.Y + Crop.Height,
				TargetIntrinsics.Fx, TargetIntrinsics.Fy, TargetIntrinsics.Cx, TargetIntrinsics.Cy, 0.1f, 10);

			// Y flip so that ReadPixels returns the image the right way up
			projMat *= new Matrix4(Vector4.UnitX, -Vector4.UnitY, Vector4.UnitZ, Vector4.UnitW);
			

			// Step 2. Modelview matrix to reproject view -----------------------------------------

			//// 2a. Fully flexible model-view matrix
			//var sourceView = TargetExtrinsics.ToMatrix4();
			//var targetView = TargetExtrinsics.ToMatrix4();
			//targetView.M14 += deltaX;

			//// undo the source view to get to world coordinates,
			//// then apply the target view matrix
			//sourceView.Invert();
			//var viewMat = targetView * sourceView;

			//// negate Y and Z rows to convert from calibration to OpenGL coordinate system
			//viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, Vector4.UnitW) * viewMat;

			//// OpenTK matrices are row-major, so transpose to get column-major data layout that OpenGL expects
			//viewMat.Transpose();

			// 2b. All operations in 2a rolled into a single matrix
			var viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, new Vector4(deltaX, 0, 0, 1));

			// 2c. original view
			//viewMat = Matrix4.Identity; // source view
			//viewMat = new Matrix4(Vector4.UnitX, -Vector4.UnitY, -Vector4.UnitZ, Vector4.UnitW); // source view (incl. calibration to OpenGL transform)


			// Step 3. Set matrices ---------------------------------------------------------------

			GL.MatrixMode(MatrixMode.Projection); GL.LoadMatrix(ref projMat);
			GL.MatrixMode(MatrixMode.Modelview);  GL.LoadMatrix(ref viewMat);
		}

		private TextureFrame[] DrawMesh(int colourSequenceNumber, int pointsSequenceNumber, float deltaX)
		{
			// allocate the output textures
			var points = new TextureFrame(Crop.Width, Crop.Height) { InternalFormat = PixelInternalFormat.Rgba32f, Wrap = TextureWrapMode.ClampToEdge, SequenceNumber = pointsSequenceNumber };
			var result = new TextureFrame(Crop.Width, Crop.Height) { InternalFormat = PixelInternalFormat.Rgba8,   Wrap = TextureWrapMode.ClampToEdge, SequenceNumber = colourSequenceNumber };
			points.Alloc();
			result.Alloc();

			SetupMeshMatrices(deltaX);

			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, FramebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, points.Texture, 0);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment1, TextureTarget.Texture2D, result.Texture, 0);
			var drawBuffers = new DrawBuffersEnum[] { DrawBuffersEnum.ColorAttachment0, DrawBuffersEnum.ColorAttachment1};
			GL.DrawBuffers(2, drawBuffers);
			Utils.CheckFboStatus();
			
			GL.ClearColor(Color.Black);
			GL.Viewport(0, 0, Crop.Width, Crop.Height);
			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

			// push the current Array Buffer state so we can restore it later
			GL.PushClientAttrib(ClientAttribMask.ClientVertexArrayBit);

			// use program and bind depth image to texture unit 0
			GL.UseProgram(MeshProgramID);
			PointCloudTexture.Bind(TextureUnit.Texture0);
			VideoTexture.Bind(TextureUnit.Texture1);
			GL.Uniform1(GL.GetUniformLocation(MeshProgramID, "shift"), Shift);
	
			// sets the vertex pointer to the currently bound array describing how the data is stored
			GL.BindBuffer(BufferTarget.ArrayBuffer, VertexBufferID);
			GL.VertexPointer(3, VertexPointerType.Float, Vector3.SizeInBytes, IntPtr.Zero);
			GL.EnableClientState(ArrayCap.VertexArray);

			// draw the triangles in the element array buffer
			GL.BindBuffer(BufferTarget.ElementArrayBuffer, IndexBufferID);
			GL.DrawElements(BeginMode.Triangles, NumIndices, DrawElementsType.UnsignedInt, IntPtr.Zero);

			// restore the state
			GL.UseProgram(0);
			PointCloudTexture.Unbind();
			GL.PopClientAttrib();

			return new TextureFrame[] { points, result };
		}

//=================================================================================================
		#endregion
	}
}
