﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Colours.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

//---- standard colour conversion functions -------------------------------------------------------

const float c0 = 1 / 12.92f;
const float c1 = 1 / 1.055f;
const float c2 = 1 / 3.0f;
const float c3 = 16.0f / 116.0f;

vec3 rgb2xyz(vec3 c)
{
	// linearise sRGB to linear RGB (range 0..1)
	float r = (c.r > .04045f) ? pow((c.r + .055f) * c1, 2.4f) : c.r * c0;
	float g = (c.g > .04045f) ? pow((c.g + .055f) * c1, 2.4f) : c.g * c0;
	float b = (c.b > .04045f) ? pow((c.b + .055f) * c1, 2.4f) : c.b * c0;

	// convert to XYZ (range 0..100)
	return 100 * vec3(r, g, b) * mat3(.4124f, .3576f, .1805f, .2126f, .7152f, .0722f, .0193f, .1192f, .9505f);
}

vec3 xyz2lab(vec3 c)
{
	// normalise with respect to white point D65
	vec3 n = c / vec3(95.047f, 100, 108.883f);
	
	// non-linear mapping
	float fx = (n.x > 0.008856) ? pow(n.x, c2) : 7.787f * n.x + c3;
	float fy = (n.y > 0.008856) ? pow(n.y, c2) : 7.787f * n.y + c3;
	float fz = (n.z > 0.008856) ? pow(n.z, c2) : 7.787f * n.z + c3;
	
	// remap values to CIELAB (range -100..100)
	return vec3(116 * fy - 16, 500 * (fx - fy), 200 * (fy - fz));
}

vec3 lab2xyz(vec3 c)
{
	// undo remapping
	float fy = (c.x + 16.0) / 116.0;
	float fx = c.y / 500.0 + fy;
	float fz = fy - c.z / 200.0;

	// non-linear remapping
	float x = (fx > 0.206897) ? fx * fx * fx : (fx - 16.0 / 116.0) / 7.787;
	float y = (fy > 0.206897) ? fy * fy * fy : (fy - 16.0 / 116.0) / 7.787;
	float z = (fz > 0.206897) ? fz * fz * fz : (fz - 16.0 / 116.0) / 7.787;

	// un-normalise with respect to white point D65
	return vec3(x, y, z) * vec3(95.047f, 100, 108.883f);
}

vec3 xyz2rgb(vec3 c)
{
	// convert CIEXYZ (0..100) to linear RGB (0..1)
	vec3 rgb = c / 100.0 * mat3(3.2406, -1.5372, -0.4986, -0.9689, 1.8758, 0.0415, 0.0557, -0.2040, 1.0570);

	// convert linear RGB (0..1) to sRGB (0..1)
	float sr = (rgb.r > 0.0031308) ? (1.055 * pow(rgb.r, 1 / 2.4) - 0.055) : 12.92 * rgb.r;
	float sg = (rgb.g > 0.0031308) ? (1.055 * pow(rgb.g, 1 / 2.4) - 0.055) : 12.92 * rgb.g;
	float sb = (rgb.b > 0.0031308) ? (1.055 * pow(rgb.b, 1 / 2.4) - 0.055) : 12.92 * rgb.b;
	
	return vec3(sr, sg, sb);
}

vec3 rgb2lab(vec3 srgb)
{
	// straightforward conversion of sRGB (range 0..1) to CIELAB (range -100..100)
	return xyz2lab(rgb2xyz(srgb));
}

vec3 lab2rgb(vec3 lab)
{
	// convert CIELAB to sRGB
	return xyz2rgb(lab2xyz(lab));
}

//---- pack/unpack LAB functions (to map between CIELAB range and 0..1) ---------------------------

vec3 pack_lab(vec3 lab)
{
	// remap values to fit within range 0..1
	return vec3(0.01 * lab.x, 0.5 * (1.0 + lab.y / 127.0), 0.5 * (1.0 + lab.z / 127.0));
}

vec3 unpack_lab(vec3 c)
{
	// map custom 0..1 range back to standard -100..100 range
	return vec3(100.0 * c.x, 254.0 * (c.y - 0.5), 254.0 * (c.z - 0.5));
}