/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: BitmapFrames.h 554 2012-08-30 17:15:28Z cr333 $
*/
#ifndef BITMAPFRAMES_H_INCLUDED
#define BITMAPFRAMES_H_INCLUDED

namespace BitmapFrames
{
	enum PixelFormat
	{
		/// <summary>Unknown pixel format.</summary>
		Unknown,

		/// <summary>Unknown pixel format.</summary>
		None = Unknown,

// data type of colour components -----------------------------------------------------------------

		/// <summary>8-bit unsigned integer.</summary>
		Byte = 1 << 0,

		/// <summary>16-bit unsigned integer.</summary>
		Short = 1 << 1,

		/// <summary>32-bit unsigned integer.</summary>
		Int = 1 << 2,

		/// <summary>32-bit single-precision floating-point number (IEEE 754).</summary>
		Single = 1 << 3,

		/// <summary>64-bit double-precision floating-point number (IEEE 754).</summary>
		Double = 1 << 4,
		
		/// <summary>Bit mask for all datatypes.</summary>
		TypeMask = Byte | Short | Int | Single | Double,

// layout of colour components --------------------------------------------------------------------

		/// <summary>Greyscale colour components.</summary>
		Grey = 1 << 10,

		/// <summary>Packed RGB colour components: RGBRGB...</summary>
		RGB = 1 << 11,

		/// <summary>Packed BGR colour components: BGRBGR...</summary>
		BGR = 1 << 12,

		/// <summary>Packed ARGB colour components: ARGBARGB...</summary>
		ARGB = 1 << 13,

		/// <summary>Packed RGBA colour components: RGBARGBA...</summary>
		RGBA = 1 << 14,

		/// <summary>Packed ABGR colour components: ABGRABGR...</summary>
		ABGR = 1 << 15,

		/// <summary>Packed BGRA colour components: BGRABGRA...</summary>
		BGRA = 1 << 16,

		/// <summary>Bit mask for all colour component layouts.</summary>
		LayoutMask = Grey | RGB | BGR | ARGB | RGBA | ABGR | BGRA,

// additional flags -------------------------------------------------------------------------------

		/// <summary>Indicates if the alpha channel, if it exists, is used.</summary>
		Alpha = 1 << 20,

// categories of pixel formats --------------------------------------------------------------------

		/// <summary>All layouts with 3 colour components.</summary>
		ThreeColours = RGB | BGR,

		/// <summary>All layouts with 4 colour components.</summary>
		FourColours = ARGB | RGBA | ABGR | BGRA,

// common formats ---------------------------------------------------------------------------------

		/// <summary>8-bit greyscale.</summary>
		Grey8 = Grey | Byte,

		/// <summary>16-bit greyscale.</summary>
		Grey16 = Grey | Short,

		/// <summary>Single-precision floating-point greyscale components.</summary>
		Float = Grey | Single,

		/// <summary>8-bit packed BGR components: BGRBGR...</summary>
		BGR24 = BGR | Byte,

		/// <summary>8-bit packed BGRX components (alpha unused): BGRXBGRX... or 0xXXRRGGBB.</summary>
		BGR32 = BGRA | Byte,
	
		/// <summary>8-bit packed BGRA components: BGRABGRA... or 0xAARRGGBB.</summary>
		BGRA32 = BGRA | Byte | Alpha,

		/// <summary>8-bit packed RGB components: RGBRGB...</summary>
		RGB24 = RGB | Byte,

		/// <summary>8-bit packed RGBX components (alpha unused): RGBXRGBX... or 0xXXBBGGRR.</summary>
		RGB32 = RGBA | Byte,
	
		/// <summary>8-bit packed RGBA components: RGBARGBA... or 0xAABBGGRR.</summary>
		RGBA32 = RGBA | Byte | Alpha,

		/// <summary>Single-precision floating-point BGR components.</summary>
		FloatBGR = BGR | Single,

		/// <summary>Single-precision floating-point BGRX components (alpha unused).</summary>
		FloatBGRX = BGRA | Single,

		/// <summary>Single-precision floating-point BGRX components.</summary>
		FloatBGRA = BGRA | Single | Alpha,

		/// <summary>Single-precision floating-point RGB components.</summary>
		FloatRGB = RGB | Single,

		/// <summary>Single-precision floating-point RGBX components (alpha unused).</summary>
		FloatRGBX = RGBA | Single,

		/// <summary>Single-precision floating-point RGBA components.</summary>
		FloatRGBA = RGBA | Single | Alpha,
	};

	struct RawHeader
	{
		//unsigned char magic[12];
		unsigned int version;
		int width, height;
		PixelFormat format;
		int stride;
	};

	bool saveRawBitmapFrame(const void* data, const RawHeader& header, const char* filename);
	bool readRawBitmapFrame(const char* filename, RawHeader& header, void** data);
}

#endif // BITMAPFRAMES_H_INCLUDED