/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: main.cpp 537 2012-04-04 15:35:42Z cr333 $
*/
#include <assert.h>
#include <stdio.h>
#include "BitmapFrames.h"

using namespace BitmapFrames;

// Example 1: read in a video frame
void readVideoFrame(const char* filename)
{
	// read in the raw data
	void* data;
	RawHeader header;
	if(!readRawBitmapFrame(filename, header, &data))
	{
		printf("Error: failed to read in '%s'.\n", filename);
		return;
	}

	// check for correct image format
	if(header.format != BGR24)
	{
		printf("Error: wrong image format; expected BGR24.\n");
		return;
	}

	// read out pixels
	unsigned char* pImage = (unsigned char*)data;
	for(int y = 0; y < header.height; y++)
	{
		unsigned char* pRow = &pImage[y * header.stride];
		for(int x = 0; x < header.width; x++)
		{
			unsigned char b = *pRow++;
			unsigned char g = *pRow++;
			unsigned char r = *pRow++;

			//printf("[%3i, %3i] -- (%3i, %3i, %3i)\n", x, y, r, g, b);
		}
	}

	// free memory
	delete data;
}

// Example 2: read in a depth frame
void readDepthFrame(const char* filename)
{
	// read in the raw data
	void* data;
	RawHeader header;
	if(!readRawBitmapFrame(filename, header, &data))
	{
		printf("Error: failed to read in '%s'.\n", filename);
		return;
	}

	// check for correct image format
	if(header.format != FloatRGBA)
	{
		printf("Error: wrong image format; expected BGR24.\n");
		return;
	}

	// read out pixels
	float* pImage = (float*)data;
	for(int y = 0; y < header.height; y++)
	{
		float* pRow = &pImage[y * header.stride / sizeof(float)];
		for(int x = 0; x < header.width; x++)
		{
			float pX = *pRow++; // positive axis points to the left
			float pY = *pRow++; // positive axis points up
			float pZ = *pRow++; // positive axis points into the screen (away from viewer)
			float pD = *pRow++; // Euclidean distance between camera centre and point (i.e. length of XYZ)

			//printf("[%3i, %3i] -- (%6.3f, %6.3f, %6.3f, %6.3f)\n", x, y, pX, pY, pZ, pD);
		}
	}

	// free memory
	delete data;
}

int main(int argc, char* argv[])
{
	// Example 1: read in a video frame
	readVideoFrame("fruitsA\\processed\\video-0100.raw");

	// Example 2: read in a depth frame
	readDepthFrame("fruitsA\\processed\\depth-0100.raw");

	printf("All done.\n");
}