﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: LightingEstimationStep1.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using RGBZcamera.Library;
using RGBZcamera.Library.AbstractedLines;
using RGBZcamera.Library.Config;
using OpenTK;
using VideoLib;
using VideoLib.Filters;
using VideoLib.Frames;
using VideoLib.OpenGL;
using VideoLib.Scheduling;
using VideoLib.Sinks;
using VideoLib.Sources;
using VideoLib.Utils;
using OpenGL = OpenTK.Graphics.OpenGL;

namespace LightingEstimationStep1
{
	class LightingEstimationStep1
	{
		static void Main(string[] args)
		{
//---- load dataset information from config file --------------------------------------------------

			var dataset    = "fruitsB";
			var dataDir    = ConfigReader.FindDataDirectory(".");
			var configFile = Path.Combine(dataDir.FullName, String.Format(@"input\{0}\_config.ini", dataset));
			var config     = new ConfigReader(configFile);

//--- do the work ---------------------------------------------------------------------------------

			CreateNormalMapsAndShadedImages(config);
			//RenderFrames(config, 653);
			ExportData(config, 50);
		}

		/// <summary>Computes normals and bilaterally filtered shading images from processed RGBZ frames.</summary>
		/// <param name="config">The config object describing the input dataset.</param>
		private static void CreateNormalMapsAndShadedImages(ConfigReader config)
		{
			Topology    topology  = Topology.MainInstance;
			Scheduler   scheduler = new ModularScheduler(topology, /* new CudaWorkQueue(), */ new OpenGLWorkQueue(), new StatusWorkQueue(), new SharedWorkQueue(4)) { Profiling = true };
			BitmapFrame.Allocator = new BuiltinAllocator();

			// load filtered distance maps and video frames from disk
			var colour = new RawSource(Path.Combine(config.OutputDir, @"processed\video-{0:D4}.raw")) { FPS = 15, FirstNumber = config.Video.FirstFrame, LastNumber = config.Video.LastFrame, Loop = false };
			var points = new RawSource(Path.Combine(config.OutputDir, @"processed\depth-{0:D4}.raw")) { FPS = 15, FirstNumber = config.Depth.FirstFrame, LastNumber = config.Depth.LastFrame, Loop = false }.OutputPins["out"];

			// copy data to textures
			var colourTex = topology.Connect(colour, new TextureInput() { InternalFormat = OpenGL::PixelInternalFormat.Rgba32f, Wrap = OpenGL::TextureWrapMode.ClampToEdge });
			var pointsTex = topology.Connect(points, new TextureInput() { InternalFormat = OpenGL::PixelInternalFormat.Rgba32f });

			// compute and save normals to disk
			var normals = new ToneRendererNode() { ToneOutput = ToneOutput.Normals };
			topology.Connect(colourTex, normals["colour"]);
			topology.Connect(pointsTex, normals["depth"]);
			topology.Connect(normals, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, new RawSink(Path.Combine(config.OutputDir, @"processed\normals-{0:D4}.raw")));
			topology.Connect(normals, new TextureOutput(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"processed\normals-{0:D4}.png")));

			// compute and save bilaterally filtered image to disk (raw and png)
			var bilateralFilter = topology.Connect(colourTex, new BilateralFilter() { SpatialSigma = 5, RangeSigma = 0.1f });
			topology.Connect(bilateralFilter, new TextureOutput() { TargetFormat = PixelFormat.FloatRGBA }, new RawSink(Path.Combine(config.OutputDir, @"processed\shading-{0:D4}.raw")));
			topology.Connect(bilateralFilter, new TextureOutput(), new ImageSequenceSink(Path.Combine(config.OutputDir, @"processed\shading-{0:D4}.png")));

			// start the scheduler
			scheduler.Setup();
			scheduler.StartWorkers(4);
			scheduler.Join();
		}

		///// <summary>Renders diffusely lit images for a 10 degree-sampled set of light directions.</summary>
		///// <param name="config">The config object describing the input dataset.</param>
		///// <param name="srFrame">The ToF frame number.</param>
		//private static void RenderFrames(ConfigReader config, int srFrame)
		//{
		//    // create a GraphicsContext
		//    INativeWindow window = new NativeWindow();
		//    OpenTK.Graphics.IGraphicsContext context = new OpenTK.Graphics.GraphicsContext(OpenTK.Graphics.GraphicsMode.Default, window.WindowInfo);
		//    context.MakeCurrent(window.WindowInfo);
		//    context.LoadAll();

		//    // make sure target directory exists
		//    var targetDir = Path.Combine(config.OutputDir, "shaded");
		//    if(!File.Exists(targetDir))
		//        Directory.CreateDirectory(targetDir);

		//    // load points and colour image
		//    int fleaFrame = srFrame - config.Depth.FirstFrame + config.Video.FirstFrame;
		//    var rawPoints = RawSource.LoadFrameFromFile(Path.Combine(config.OutputDir, string.Format(@"processed\depth-{0:D4}.raw", srFrame)));
		//    var rawColour = RawSource.LoadFrameFromFile(Path.Combine(config.OutputDir, string.Format(@"processed\shading-{0:D4}.raw", fleaFrame)));

		//    // convert to textures
		//    var texPoints = new TextureFrame(rawPoints.Width, rawPoints.Height) { InternalFormat = OpenGL::PixelInternalFormat.Rgba32f };
		//    var texColour = new TextureFrame(rawColour.Width, rawColour.Height) { InternalFormat = OpenGL::PixelInternalFormat.Rgba32f };
		//    texPoints.Texture.Create(rawPoints);
		//    texColour.Texture.Create(rawColour);
		//    var inputTextures = new Dictionary<string, TextureFrame> { { "colour", texColour }, { "depth", texPoints } };

		//    var renderer = new ToneRendererNode() { ToneOutput = ToneOutput.Diffuse, Ambient = 0, Diffuse = 1, Specular = 0, IsPosition = false };
		//    renderer.Init();

		//    for (int azimuth = 0; azimuth < 360; azimuth += 10)
		//    {
		//        var phi = azimuth / 180.0f * (float)Math.PI;
		//        for (int zenith = 0; zenith <= 90; zenith += 10)
		//        {
		//            var theta = zenith / 180.0f * (float)Math.PI;

		//            var x = Math.Cos(phi) * Math.Cos(theta);
		//            var y = Math.Sin(phi) * Math.Cos(theta);
		//            var z = -Math.Sin(theta);
		//            renderer.LightDirection = new Vector3((float)x, (float)y, (float)z);

		//            Console.WriteLine("Rendering azimuth {0:D3}, zenith {1:D3} ...", azimuth, zenith);
		//            var result = renderer.Draw(inputTextures);
		//            var bitmap = result.Texture.ReadFrame(PixelFormat.BGR24).Bitmap;
		//            result.RefDec();

		//            var path = Path.Combine(config.OutputDir, string.Format(@"shaded\zenith-{1:D3}-azimuth-{0:D3}.png", azimuth, zenith));
		//            bitmap.Save(path);
		//        }
		//    }

		//    texColour.RefDec();
		//    texPoints.RefDec();

		//    GC.KeepAlive(window);
		//}

		/// <summary>Writes a text-based file containing per-pixel colour and normals for the lighting estimation.</summary>
		/// <param name="config">The config object describing the input dataset.</param>
		/// <param name="srFrame">The video frame number.</param>
		private unsafe static void ExportData(ConfigReader config, int videoFrame)
		{
			// load points, normals and shading image
			int srFrame = videoFrame - config.Video.FirstFrame + config.Depth.FirstFrame;
			var rawPoints  = RawSource.LoadFrameFromFile(Path.Combine(config.OutputDir, string.Format(@"processed\depth-{0:D4}.raw", srFrame)));
			var rawNormals = RawSource.LoadFrameFromFile(Path.Combine(config.OutputDir, string.Format(@"processed\normals-{0:D4}.raw", videoFrame)));
			var rawShading = RawSource.LoadFrameFromFile(Path.Combine(config.OutputDir, string.Format(@"processed\shading-{0:D4}.raw", videoFrame)));

			// also save points as OFF
			var export = new OffExportSink(Path.Combine(config.OutputDir, string.Format(@"processed\frame-{0:D4}.off", videoFrame)));
			export.SaveFrame(rawPoints);

			// check format and size
			var allFrames = new Dictionary<string, BitmapFrame> { { "points", rawPoints }, {"normals", rawNormals}, {"shading", rawShading}};
			Node.AssertFrameProperties(allFrames, Node.FrameProps.Format | Node.FrameProps.Size | Node.FrameProps.Stride);
			Node.AssertPixelFormat(allFrames, PixelFormat.FloatRGBA);

			// write one point per line: x y n_x n_y n_z r g b
			using (var writer = new StreamWriter(Path.Combine(config.OutputDir, string.Format(@"processed\frame-{0:D4}.txt", videoFrame))))
			{
				fixed (byte* bytePoints = &rawPoints.Data[0], byteNormals = &rawNormals.Data[0], byteShading = &rawShading.Data[0])
				{
					float* points = (float*)bytePoints;
					float* normals = (float*)byteNormals;
					float* shading = (float*)byteShading;

					for (int y = 0; y < rawPoints.Height; y++)
					{
						for (int x = 0; x < rawPoints.Width; x++)
						{
							int index = 4 * y * rawPoints.Width + 4 * x;

							// skip NaN normals
							if (float.IsNaN(normals[index + 0]) || float.IsNaN(normals[index + 0]) || float.IsNaN(normals[index + 0]))
								continue;

							writer.Write(string.Format("{0} {1} ", x, y));
							//writer.Write(string.Format("{0:F4} {1:F4} {2:F4} ", points[index + 0], points[index + 1], points[index + 2]));
							writer.Write(string.Format("{0:F4} {1:F4} {2:F4} ", normals[index + 0], normals[index + 1], normals[index + 2]));
							writer.Write(string.Format("{0:F4} {1:F4} {2:F4}", shading[index + 0], shading[index + 1], shading[index + 2]));
							writer.WriteLine();
						}
					}
				}
			}
		}
	}
}
