// LibAvWrap -- .NET wrapper for the libav* libraries
// --------------------------------------------------
// By Christian Richardt <christian@richardt.name>
// (based on AvDn 0.7, www.informikon.com/tutorials/)
//
// $Id: Frame.h 436 2010-10-15 16:26:38Z cr333 $

#pragma once

using namespace System;
using namespace System::Text;
using namespace System::Runtime::InteropServices;

extern "C" {
#include "libswscale/swscale.h"
}

#include "Enums.h"
#include "Wrapper.h"

///<summary>Wrapper for some lavc and lavf structures.</summary>
namespace LibAvWrap
{
	///<summary>Wrapper for the AVFrame structure.</summary>
	public ref class LibAvFrame : Wrapper<AVFrame>
	{
		int width;
		int height;
		LibAvWrap::PixelFormat pixelFormat;

		bool needToFree;
		bool needToFreeData;

// constructors and destructor --------------------------------------------------------------------

	public:
		///<summary>Allocate a new frame.</summary>
		LibAvFrame() : Wrapper<AVFrame>()
		{
			pointer = avcodec_alloc_frame();
			needToFree = true;
			needToFreeData = false;

			width  = -1;
			height = -1;
			pixelFormat = LibAvWrap::PixelFormat::None;
		};

		///<summary>Allocate a new frame, including data.</summary>
		LibAvFrame(LibAvWrap::PixelFormat pixelFormat, int width, int height) : Wrapper<AVFrame>()
		{
			// create a new frame
			pointer = avcodec_alloc_frame();
			if (!pointer) throw gcnew Exception("Error: avcodec_alloc_frame failed.");

			// create image buffer
			int size = avpicture_get_size((::PixelFormat)pixelFormat, width, height);
			uint8_t* picture_buf = (uint8_t*)av_malloc(size);
			if (!picture_buf) {
				av_free(pointer);
				throw gcnew Exception("Error: av_malloc failed.");
			}
			avpicture_fill((AVPicture*)pointer, picture_buf, (::PixelFormat)pixelFormat, width, height);
			needToFree = true;
			needToFreeData = true;

			this->width       = width;
			this->height      = height;
			this->pixelFormat = pixelFormat;
		}

		LibAvFrame(array<System::Byte>^ bytes, LibAvWrap::PixelFormat pixelFormat, int width, int stride, int height) : Wrapper<AVFrame>()
		{
			// create a new frame
			pointer = avcodec_alloc_frame();
			if (!pointer) throw gcnew Exception("Error: avcodec_alloc_frame failed.");

			// create image buffer
			uint8_t* picture_buf = (uint8_t*)av_malloc(stride * height);
			if (!picture_buf) {
				av_free(pointer);
				throw gcnew Exception("Error: av_malloc failed.");
			}
			avpicture_fill((AVPicture*)pointer, picture_buf, (::PixelFormat)pixelFormat, width, height);
			needToFree = true;
			needToFreeData = true;

			pointer->linesize[0] = stride;
			Marshal::Copy(bytes, 0, IntPtr((void*)(pointer->data[0])), stride * height);
			
			this->width       = width;
			this->height      = height;
			this->pixelFormat = pixelFormat;
		}

	internal:
		LibAvFrame(AVFrame* frame, LibAvWrap::PixelFormat pixelFormat, int width, int height) : Wrapper<AVFrame>(frame)
		{
			needToFree = false;
			needToFreeData = false;

			this->width       = width;
			this->height      = height;
			this->pixelFormat = pixelFormat;
		}

	public:
		///<summary>Frees and destructs a frame.</summary>
		~LibAvFrame()
		{
			if(needToFreeData && pointer->data[0])
			{
				av_free(pointer->data[0]);
			}
			
			if(needToFree && pointer)
			{
				av_free(pointer);
				pointer = NULL;
			}
		}

	protected:
		///<summary>Finaliser, called by garbage collector.</summary>
		!LibAvFrame()
		{
			if(needToFreeData && pointer->data[0])
			{
				av_free(pointer->data[0]);
			}
			
			if(needToFree && pointer)
			{
				av_free(pointer);
				pointer = NULL;
			}
		}

// member functions -------------------------------------------------------------------------------
	public:
		///<summary>Returns the linesize of the first plane.</summary>
		int GetLinesize0() { return pointer->linesize[0]; }

		///<summary>Returns the linesize of the second plane.</summary>
		int GetLinesize1() { return pointer->linesize[1]; }

		///<summary>Returns the linesize of the third plane.</summary>
		int GetLinesize2() { return pointer->linesize[2]; }

		///<summary>Returns the linesize of the fourth plane.</summary>
		int GetLinesize3() { return pointer->linesize[3]; }

		///<summary>Returns a pointer to the first plane.</summary>
		void* GetData0() { return (void*)pointer->data[0]; }

		///<summary>Returns a pointer to the second plane.</summary>
		void* GetData1() { return (void*)pointer->data[1]; }

		///<summary>Returns a pointer to the third plane.</summary>
		void* GetData2() { return (void*)pointer->data[2]; }

		///<summary>Returns a pointer to the fourth plane.</summary>
		void* GetData3() { return (void*)pointer->data[3]; }

		///<summary>Fill in the AVPicture fields.</summary>
		void FillPicStruct(IntPtr bufPtr, LibAvWrap::PixelFormat fmt, int w, int h)
		{
			uint8_t* buf = (uint8_t*)(void*)bufPtr;
			avpicture_fill((AVPicture*)pointer, buf, (::PixelFormat)fmt, w, h);
			needToFree = false;
			// TODO: error handling
		}

		///<summary>Converts a frame to a different format using sws_scale.</summary>
		///<param name="dst">Target frame. Needs to have same size as the source image.</param>
		void ConvertTo(LibAvFrame^ dst)
		{
			// use SWS_POINT for 1:1 copy (no scaling involved here)
			Scale(this, pixelFormat, width, height,
				  dst, dst->PixelFormat, width, height,
				  SwsFlags::Point);
		}

		///<summary>Converts a frame to a different PixelFormat using sws_scale.</summary>
		static void Convert(LibAvFrame^ src, LibAvWrap::PixelFormat srcFormat, LibAvFrame^ dst, LibAvWrap::PixelFormat dstFormat, int width, int height)
		{
			// use SWS_POINT for 1:1 copy (no scaling involved here)
			Scale(src, srcFormat, width, height,
				  dst, dstFormat, width, height,
				  SwsFlags::Point | SwsFlags::PrintInfo);
		}

		///<summary>Scales frames and/or converts between PixelFormats using sws_scale.</summary>
		static void Scale(LibAvFrame^ src, LibAvWrap::PixelFormat srcFormat, int srcWidth, int srcHeight,
						  LibAvFrame^ dst, LibAvWrap::PixelFormat dstFormat, int dstWidth, int dstHeight,
						  SwsFlags flags)
		{
			// create a scaling context
			SwsContext* context = sws_getContext(
				srcWidth, srcHeight, (::PixelFormat)srcFormat,
				dstWidth, dstHeight, (::PixelFormat)dstFormat,
				(int)flags, NULL, NULL, NULL);
			// note: SWS_FAST_BILINEAR crashed the sink at 2009-03-06

			if (context == NULL)
				throw gcnew Exception("Cannot initialize the conversion context.");

			try
			{
				// apply the scaling
				sws_scale(context, src->Pointer->data, src->Pointer->linesize, 0, srcHeight,
								   dst->Pointer->data, dst->Pointer->linesize);

				// clean up
				sws_freeContext(context);
			}
			catch(Exception^ e)
			{
				throw gcnew Exception(
					String::Format("Input image: {0}x{1}, {2}, pitch {3}\nOutput image: {4}x{5}, {6}, {7}\nflags: {8}", src->Width, src->Height, src->PixelFormat, src->Stride, dst->Width, dst->Height, dst->PixelFormat, dst->Stride, flags),
					e);
			}
		}

		///<summary>Scales frames and/or converts between PixelFormats using sws_scale.</summary>
		static void Scale(LibAvFrame^ src, LibAvFrame^ dst, SwsFlags flags)
		{
			// create a scaling context
			SwsContext* context = sws_getContext(
				src->Width, src->Height, (::PixelFormat)src->PixelFormat,
				dst->Width, dst->Height, (::PixelFormat)dst->PixelFormat,
				(int)flags, NULL, NULL, NULL);
			// note: SWS_FAST_BILINEAR crashed the sink at 2009-03-06

			if (context == NULL)
				throw gcnew Exception("Cannot initialize the conversion context.");

			try
			{
				// apply the scaling
				sws_scale(context, src->Pointer->data, src->Pointer->linesize, 0, src->Height,
							   dst->Pointer->data, dst->Pointer->linesize);
				
				// clean up
				sws_freeContext(context);
			}
			catch(Exception^ e)
			{
				throw gcnew Exception(
					String::Format("Input image: {0}x{1}, {2}, pitch {3}\nOutput image: {4}x{5}, {6}, {7}\nflags: {8}", src->Width, src->Height, src->PixelFormat, src->Stride, dst->Width, dst->Height, dst->PixelFormat, dst->Stride, flags),
					e);
			}
		}

		// TODO: implement cached scaling/conversion which reuses SwsContexts
		
// properties -------------------------------------------------------------------------------------

		///<summary>Width of this frame.</summary>
		property IntPtr Scan0 { IntPtr get() { return IntPtr((void*)pointer->data[0]); } }
		
		///<summary>Width of this frame.</summary>
		property int Width { int get() { return width; } }
		
		///<summary>Stride (number of bytes per row) of this frame.</summary>
		property int Stride { int get() { return GetLinesize0(); } }
		
		///<summary>Height of this frame.</summary>
		property int Height { int get() { return height; } }
		
		///<summary>Pixel format of this frame.</summary>
		property LibAvWrap::PixelFormat PixelFormat { LibAvWrap::PixelFormat get() { return pixelFormat; } }
	};

} // namespace LibAvWrap