﻿// $Id: FboNode.cs 527 2011-04-07 20:16:55Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace VideoLib.OpenGL
{
	/// <summary>A node that renders into a frame buffer object (FBO).</summary>
	public abstract class FboNode : Node, IOpenGLNode
	{
		/// <summary>Initialises a new instance of the <see cref="FboNode"/> class.</summary>
		public FboNode()
		{
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get inputs
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);

			// draw quad
			var result = Draw(textures);
			PushFrame(result);

			// clean up
			result.RefDec();
			textures.RefDec();
		}

//=================================================================================================
		#region Rendering functions

		/// <summary>Name of the FBO used for rendering.</summary>
		protected int framebufferID;

		/// <summary>Name of the shader program used for rendering.</summary>
		protected int programID;

		/// <summary>Defines a map of textures from pin name to uniform name and texture unit.</summary>
		protected List<TextureBinding> textureMap;

		/// <summary>List of all uniforms found using <see cref="UniformAttribute"/>.</summary>
		protected IEnumerable<UniformInfo> uniforms;

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// generate FBO which is set up in the draw function
			GL.GenFramebuffers(1, out framebufferID);

			// if texture map is defined, bind all samplers
			if (textureMap != null && programID != 0)
			{
				// sort bindings by descending texture unit (so that #0 is last)
				textureMap = textureMap.OrderByDescending(x => x.TextureUnit).ToList();

				// set texture bindings to uniform names
				GL.UseProgram(programID);
				foreach (var binding in textureMap)
				{
					var location = GL.GetUniformLocation(programID, binding.Uniform);
					if (location == -1)
					{
						Console.WriteLine("Warning: Uniform '{0}' not found in '{1}'.", binding.Uniform, this);
						continue;
					}
					GL.Uniform1(location, binding.Unit);
				}
				GL.UseProgram(0);
			}
			else
			{
				// empty list is better than null (reduces extra checks later)
				textureMap = new List<TextureBinding>();
			}
		}
		
		/// <summary>Draws to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public abstract TextureFrame Draw(Dictionary<string, TextureFrame> inputs);

		/// <summary>Draws to the output texture via FBO, using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <param name="output">The texture to draw to.</param>
		protected virtual void Draw(Dictionary<string, TextureFrame> inputs, Texture output)
		{
			UpdateUniforms();
			BindTextures(inputs);
			DrawScreenQuad(output);
			UnbindTextures();
		}

		/// <summary>Binds textures (in descending order of texture unit) using the defined texture map.</summary>
		/// <param name="textures">Input texture frames.</param>
		protected void BindTextures(Dictionary<string, TextureFrame> textures)
		{
			// go through all texture bindings with pin name
			foreach (var binding in textureMap.Where(x => x.Pin != null))
				if (textures.ContainsKey(binding.Pin))
					textures[binding.Pin].Bind(binding.TextureUnit);
		}

		/// <summary>Updates the uniforms (only if their values change).</summary>
		protected void UpdateUniforms()
		{
			GL.UseProgram(programID);

			// initialise list of uniforms on first invocation
			if(uniforms == null)
				uniforms = UniformReflector.EnumerateUniforms(this, programID);

			// update all uniforms
			foreach (var uniform in uniforms)
				uniform.UpdateUniform();
		}

		/// <summary>Draws a screen quad into the FBO.</summary>
		/// <param name="target">The target texture to draw to via the FBO.</param>
		protected void DrawScreenQuad(Texture target)
		{
			// set up FBO
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, framebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, target, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			GL.Disable(EnableCap.DepthTest);
			GL.UseProgram(programID);

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();

			GL.Viewport(0, 0, target.Width, target.Height);

			// draw screen filling quad with appropriate texture coordinates
			GL.Begin(BeginMode.Quads);
			GL.TexCoord2(0, 0); GL.Vertex2(-1, -1);
			GL.TexCoord2(1, 0); GL.Vertex2( 1, -1);
			GL.TexCoord2(1, 1); GL.Vertex2( 1,  1);
			GL.TexCoord2(0, 1); GL.Vertex2(-1,  1);
			GL.End();

			GL.UseProgram(0);
			GL.Enable(EnableCap.DepthTest);
			GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
		}

		/// <summary>Unbinds textures from all mapped texture units.</summary>
		protected void UnbindTextures()
		{
			// go through all texture bindings (not just the ones with pin name)
			foreach (var binding in textureMap)
			{
				GL.ActiveTexture(binding.TextureUnit);
				GL.BindTexture(TextureTarget.Texture2D, 0);
			}
		}

//=================================================================================================
		#endregion // Rendering functions
	}
}
