﻿// $Id: VectorConverter.cs 515 2010-12-17 12:48:50Z cr333 $
using System;
using System.ComponentModel;
using System.Linq;
using OpenTK;

namespace VideoLib.OpenGL
{
	/// <summary>Type converter for OpenTK vectors.</summary>
	public class VectorConverter : TypeConverter
	{
		/// <summary>Returns whether this converter can convert the object to the specified type, using the specified context.</summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
		/// <param name="destinationType">A <see cref="T:System.Type"/> that represents the type you want to convert to.</param>
		/// <returns>true if this converter can perform the conversion; otherwise, false.</returns>
		public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
		{
			if (destinationType == typeof(Vector2)) return true;
			if (destinationType == typeof(Vector3)) return true;
			if (destinationType == typeof(Vector4)) return true;

			return base.CanConvertTo(context, destinationType);
		}

		/// <summary>Converts the given value object to the specified type, using the specified context and culture information.</summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
		/// <param name="culture">A <see cref="T:System.Globalization.CultureInfo"/>. If null is passed, the current culture is assumed.</param>
		/// <param name="value">The <see cref="T:System.Object"/> to convert.</param>
		/// <param name="destinationType">The <see cref="T:System.Type"/> to convert the <paramref name="value"/> parameter to.</param>
		/// <returns>An <see cref="T:System.Object"/> that represents the converted value.</returns>
		/// <exception cref="T:System.ArgumentNullException">The <paramref name="destinationType"/> parameter is null.</exception>
		/// <exception cref="T:System.NotSupportedException">The conversion cannot be performed.</exception>
		public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
		{
			if (value is Vector2 && destinationType == typeof(string)) return value.ToString();
			if (value is Vector3 && destinationType == typeof(string)) return value.ToString();
			if (value is Vector4 && destinationType == typeof(string)) return value.ToString();

			return base.ConvertTo(context, culture, value, destinationType);
		}

		/// <summary>Returns whether this converter can convert an object of the given type to the type of this converter, using the specified context.</summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
		/// <param name="sourceType">A <see cref="T:System.Type"/> that represents the type you want to convert from.</param>
		/// <returns>true if this converter can perform the conversion; otherwise, false.</returns>
		public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
		{
			if (sourceType == typeof(string))
				return true;

			return base.CanConvertFrom(context, sourceType);
		}

		/// <summary>Converts the given object to the type of this converter, using the specified context and culture information.</summary>
		/// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
		/// <param name="culture">The <see cref="T:System.Globalization.CultureInfo"/> to use as the current culture.</param>
		/// <param name="value">The <see cref="T:System.Object"/> to convert.</param>
		/// <returns>An <see cref="T:System.Object"/> that represents the converted value.</returns>
		/// <exception cref="T:System.NotSupportedException">The conversion cannot be performed.</exception>
		public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
		{
			if (value is string)
			{
				try
				{
					var components = ((string)value).Trim('(', ')', '[', ']').Split(',', ';', ' ').Where(x => x != "").Select(x => float.Parse(x.Trim())).ToArray();
					switch (components.Length)
					{
						case 2: return new Vector2(components[0], components[1]);
						case 3: return new Vector3(components[0], components[1], components[2]);
						case 4: return new Vector4(components[0], components[1], components[2], components[3]);
					}
				}
				catch
				{
					throw new ArgumentException(string.Format("Cannot convert '{0}' to OpenTK.Vector[234].", value));
				}
			}

			return base.ConvertFrom(context, culture, value);
		}
	}
}
