﻿// $Id: KinectSource.cs 544 2011-09-26 17:59:08Z cr333 $
using System;
using OpenNI;
using VideoLib.Frames;

namespace VideoLib.OpenNI
{
	/// <summary>Captures live stereo video from a Kinect stereo webcam.</summary>
	public class KinectSource : Node, IDisposable
	{
		/// <summary>The sequence number of the most recently captured frame.</summary>
		public int SequenceNumber { get; protected set; }

		/// <summary>The video generators captured by this Kinect.</summary>
		public Streams Streams { get; protected set; }

		public bool CaptureImage { get { return (Streams & Streams.Image) == Streams.Image; } }
		public bool CaptureDepth { get { return (Streams & Streams.Depth) == Streams.Depth; } }
		public bool CaptureIR    { get { return (Streams & Streams.IR   ) == Streams.IR;    } }

		public ImageModes ImageMode { get; set; }
		public DepthModes DepthMode { get; set; }
		public IRModes    IRMode    { get; set; }

		public SyncMode SyncMode { get; set; }


		/// <summary>Constructs a video source capturing image and depth from a Microsoft Kinect sensor.</summary>
		public KinectSource() : this(Streams.ImageAndDepth)
		{
		}

		/// <summary>Constructs a video source capturing from a Microsoft Kinect sensor.</summary>
		public KinectSource(Streams streams) : base()
		{
			// default settings
			ImageMode      = ImageModes.VGA_30fps;
			DepthMode      = DepthModes.VGA_30fps;
			IRMode         = IRModes.VGA_30fps;
			SyncMode       = SyncMode.DropFrames;
			SequenceNumber = 0;

			Streams = streams;
			if (CaptureImage) AddOutputPin("image");
			if (CaptureDepth) AddOutputPin("depth");
			if (CaptureIR)    AddOutputPin("ir");
		}

		private long lastFrameTimestamp = 0;

		/// <summary>Outputs a pair of depth and image frames.</summary>
		public unsafe override void Process()
		{
			while(true)
			{
				// wait for frame(s)
				if(SyncMode == SyncMode.WaitAnd) context.WaitAndUpdateAll();
				else                             context.WaitAnyUpdateAll(); // WaitAny and DropFrames
				//context.WaitOneUpdateAll(depth);
				//context.WaitOneUpdateAll(image);

				// 'soft'-synchronise multiple streams
				if (SyncMode == SyncMode.DropFrames && CaptureDepth && (CaptureImage || CaptureIR))
				{
					if (depth == null) throw new Exception("Depth generator not initialised.");

					long timestamp = 0;
					float targetFps = 0;

					if (CaptureImage)
					{
						if (image == null) throw new Exception("Image generator not initialised.");
						
						// drop duplicate image frame
						if (image.Timestamp <= lastFrameTimestamp)
							continue;

						timestamp = image.Timestamp;
						targetFps = image.MapOutputMode.FPS;
					}

					if (CaptureIR)
					{
						if (ir == null) throw new Exception("IR generator not initialised.");

						// drop duplicate IR frame
						if (ir.Timestamp <= lastFrameTimestamp)
							continue;

						timestamp = ir.Timestamp;
						targetFps = ir.MapOutputMode.FPS;
					}

					// Scumbag Kinect: ask for 15 fps, get 10 fps + bad synchronisation (hence fudge factor 2/3).
					if (targetFps == 15) targetFps = 10 * (2 / 3f);

					// drop frame if their timestamps differ too much
					var diff = (timestamp - depth.Timestamp) / 1.0e6;
					if (Math.Abs(diff) > 0.5 / targetFps)
					{
						continue;
					}
					else
					{
						lastFrameTimestamp = timestamp;
					}
				}

				if (CaptureImage)
				{
					if (image == null)
						throw new Exception("Image generator not initialised.");

					var imageMD = image.GetMetaData();
					var imageFrame = new BitmapFrame(image.ImageMapPtr, imageMD.XRes, 3 * imageMD.XRes, imageMD.YRes, VideoLib.Frames.PixelFormat.RGB24) { SequenceNumber = SequenceNumber };
					PushFrame("image", imageFrame);
				} 
				
				if (CaptureDepth)
				{
					if (depth == null)
						throw new Exception("Depth generator not initialised.");

					var depthMD = depth.GetMetaData();
					var depthFrame = new BitmapFrame(depth.DepthMapPtr, depthMD.XRes, sizeof(ushort) * depthMD.XRes, depthMD.YRes, VideoLib.Frames.PixelFormat.Grey16) { SequenceNumber = SequenceNumber };
					PushFrame("depth", depthFrame);
				}

				if (CaptureIR)
				{
					if (ir == null)
						throw new Exception("IR generator not initialised.");

					var irMD = ir.GetMetaData();
					var irFrame = new BitmapFrame(ir.IRMapPtr, irMD.XRes, sizeof(ushort) * irMD.XRes, irMD.YRes, VideoLib.Frames.PixelFormat.Grey16) { SequenceNumber = SequenceNumber };
					PushFrame("ir", irFrame);
				}

				SequenceNumber++;
				break; // leave infinite loop
			}
		}

		private Context context;
		private DepthGenerator depth;
		private ImageGenerator image;
		private IRGenerator ir;

		/// <summary>Initialises OpenNI for capturing.</summary>
		public override void Init()
		{
			// create context from scratch
			context = new Context();

			if (CaptureImage)
			{
				image = new ImageGenerator(context);
				image.MapOutputMode = new MapOutputMode()
				{
					XRes = GetWidth( (Modes)ImageMode),
					YRes = GetHeight((Modes)ImageMode),
					FPS  = GetFps(   (Modes)ImageMode),
				};
			}

			if (CaptureDepth)
			{
				depth = new DepthGenerator(context);
				depth.MapOutputMode = new MapOutputMode()
				{
					XRes = GetWidth( (Modes)DepthMode),
					YRes = GetHeight((Modes)DepthMode),
					FPS  = GetFps(   (Modes)DepthMode),
				};
			}

			if (CaptureIR)
			{
				ir = new IRGenerator(context);
				ir.MapOutputMode = new MapOutputMode()
				{
					XRes = GetWidth( (Modes)IRMode),
					YRes = GetHeight((Modes)IRMode),
					FPS  = GetFps(   (Modes)IRMode),
				};
			}

			context.StartGeneratingAll();
		}

		protected void CleanUp()
		{
			if (context != null)
			{
				context.Release();
				context = null;
			}
		}

		protected override void Dispose(bool disposing)
		{
			CleanUp();
			base.Dispose(disposing);
		}

		#region Utility functions

		private int GetWidth(Modes mode)
		{
			if ((mode & Modes.QVGA) == Modes.QVGA) return  320;
			if ((mode & Modes.VGA)  == Modes.VGA)  return  640;
			if ((mode & Modes.SXGA) == Modes.SXGA) return 1280;
			if ((mode & Modes.UXGA) == Modes.UXGA) return 1600;
			return 0;
		}

		private int GetHeight(Modes mode)
		{
			if ((mode & Modes.QVGA) == Modes.QVGA) return  240;
			if ((mode & Modes.VGA)  == Modes.VGA)  return  480;
			if ((mode & Modes.SXGA) == Modes.SXGA) return 1024;
			if ((mode & Modes.UXGA) == Modes.UXGA) return 1200;
			return 0;
		}

		private int GetFps(Modes mode)
		{
			if ((mode & Modes.Fps15) == Modes.Fps15) return 15;
			if ((mode & Modes.Fps25) == Modes.Fps25) return 25;
			if ((mode & Modes.Fps30) == Modes.Fps30) return 30;
			if ((mode & Modes.Fps60) == Modes.Fps60) return 60;
			return 0;
		}

		#endregion Utility functions
	}
}