﻿// $Id: PixelFormat.cs 336 2010-08-17 18:07:55Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Imaging = System.Drawing.Imaging;

namespace VideoLib.Frames
{
	/// <summary>
	/// List of supported pixel formats.
	/// </summary>
	public enum PixelFormat
	{
		/// <summary>Unknown pixel format.</summary>
		Unknown,

		/// <summary>Unknown pixel format.</summary>
		None = Unknown,

// data type of colour components -----------------------------------------------------------------

		/// <summary>8-bit unsigned integer.</summary>
		Byte = 1 << 0,

		/// <summary>16-bit unsigned integer.</summary>
		Short = 1 << 1,

		/// <summary>32-bit unsigned integer.</summary>
		Int = 1 << 2,

		/// <summary>32-bit single-precision floating-point number (IEEE 754).</summary>
		Single = 1 << 3,

		/// <summary>64-bit double-precision floating-point number (IEEE 754).</summary>
		Double = 1 << 4,
		
		/// <summary>Bit mask for all datatypes.</summary>
		TypeMask = Byte | Short | Int | Single | Double,

// layout of colour components --------------------------------------------------------------------

		/// <summary>Greyscale colour components.</summary>
		Grey = 1 << 10,

		/// <summary>Packed RGB colour components: RGBRGB...</summary>
		RGB = 1 << 11,

		/// <summary>Packed BGR colour components: BGRBGR...</summary>
		BGR = 1 << 12,

		/// <summary>Packed ARGB colour components: ARGBARGB...</summary>
		ARGB = 1 << 13,

		/// <summary>Packed RGBA colour components: RGBARGBA...</summary>
		RGBA = 1 << 14,

		/// <summary>Packed ABGR colour components: ABGRABGR...</summary>
		ABGR = 1 << 15,

		/// <summary>Packed BGRA colour components: BGRABGRA...</summary>
		BGRA = 1 << 16,

		/// <summary>Bit mask for all colour component layouts.</summary>
		LayoutMask = Grey | RGB | BGR | ARGB | RGBA | ABGR | BGRA,

// additional flags -------------------------------------------------------------------------------

		/// <summary>Indicates if the alpha channel, if it exists, is used.</summary>
		Alpha = 1 << 20,

// categories of pixel formats --------------------------------------------------------------------

		/// <summary>All layouts with 3 colour components.</summary>
		ThreeColours = RGB | BGR,

		/// <summary>All layouts with 4 colour components.</summary>
		FourColours = ARGB | RGBA | ABGR | BGRA,

// common formats ---------------------------------------------------------------------------------

		/// <summary>8-bit greyscale.</summary>
		Grey8 = Grey | Byte,

		/// <summary>16-bit greyscale.</summary>
		Grey16 = Grey | Short,

		/// <summary>Single-precision floating-point greyscale components.</summary>
		Float = Grey | Single,

		/// <summary>8-bit packed BGR components: BGRBGR...</summary>
		BGR24 = BGR | Byte,

		/// <summary>8-bit packed BGRX components (alpha unused): BGRXBGRX... or 0xXXRRGGBB.</summary>
		BGR32 = BGRA | Byte,
	
		/// <summary>8-bit packed BGRA components: BGRABGRA... or 0xAARRGGBB.</summary>
		BGRA32 = BGRA | Byte | Alpha,

		/// <summary>8-bit packed RGB components: RGBRGB...</summary>
		RGB24 = RGB | Byte,

		/// <summary>8-bit packed RGBX components (alpha unused): RGBXRGBX... or 0xXXBBGGRR.</summary>
		RGB32 = RGBA | Byte,
	
		/// <summary>8-bit packed RGBA components: RGBARGBA... or 0xAABBGGRR.</summary>
		RGBA32 = RGBA | Byte | Alpha,

		/// <summary>Single-precision floating-point BGR components.</summary>
		FloatBGR = BGR | Single,

		/// <summary>Single-precision floating-point BGRX components (alpha unused).</summary>
		FloatBGRX = BGRA | Single,

		/// <summary>Single-precision floating-point BGRX components.</summary>
		FloatBGRA = BGRA | Single | Alpha,

		/// <summary>Single-precision floating-point RGB components.</summary>
		FloatRGB = RGB | Single,

		/// <summary>Single-precision floating-point RGBX components (alpha unused).</summary>
		FloatRGBX = RGBA | Single,

		/// <summary>Single-precision floating-point RGBA components.</summary>
		FloatRGBA = RGBA | Single | Alpha,
	}

	/// <summary>Includes some format translation utilities as extension methods.</summary>
	public static class PixelFormatUtils
	{
		/// <summary>
		/// Converts the pixel format from .NET to own.
		/// </summary>
		/// <param name="pixelFormat">The pixel format.</param>
		public static PixelFormat ToVideoLibPixelFormat(this Imaging.PixelFormat pixelFormat)
		{
			switch (pixelFormat)
			{
				case Imaging.PixelFormat.Format16bppGrayScale: return PixelFormat.Grey16;
				case Imaging.PixelFormat.Format24bppRgb: return PixelFormat.BGR24;
				case Imaging.PixelFormat.Format32bppRgb: return PixelFormat.BGR32;
				case Imaging.PixelFormat.Format32bppArgb: return PixelFormat.BGRA32;
				// there is no simple 8-bit greyscale, only 8-bit palette-based => ignore it
				// there is no float-based format => ignore it
				default: return PixelFormat.Unknown;
			}
		}

		/// <summary>
		/// Gets the bitmap's pixel format as a VideoLib pixel format.
		/// </summary>
		/// <param name="bitmap">The bitmap.</param>
		public static PixelFormat VideoLibPixelFormat(this System.Drawing.Bitmap bitmap)
		{
			return bitmap.PixelFormat.ToVideoLibPixelFormat();
		}

		/// <summary>
		/// Converts the pixel format from own to .NET.
		/// </summary>
		/// <param name="formatType">The pixel format.</param>
		public static Imaging.PixelFormat ToImagingPixelFormat(this PixelFormat formatType)
		{
			switch (formatType)
			{
				case PixelFormat.Unknown: return Imaging.PixelFormat.Undefined;
				case PixelFormat.Grey8:   return Imaging.PixelFormat.Format8bppIndexed;
				case PixelFormat.Grey16:  return Imaging.PixelFormat.Format16bppGrayScale;
				case PixelFormat.BGR24:   return Imaging.PixelFormat.Format24bppRgb;
				case PixelFormat.BGR32:   return Imaging.PixelFormat.Format32bppRgb;
				case PixelFormat.BGRA32:  return Imaging.PixelFormat.Format32bppArgb;
				default: return Imaging.PixelFormat.Undefined;
			}
		}
		
		/// <summary>Returns the number of bytes per colour component.</summary>
		/// <param name="format">The pixel format.</param>
		public static int BytesPerComponent(this PixelFormat format)
		{
			switch (format & PixelFormat.TypeMask)
			{
				case PixelFormat.Byte:   return 1;
				case PixelFormat.Short:  return 2;
				case PixelFormat.Int:    return 4;
				case PixelFormat.Single: return 4;
				case PixelFormat.Double: return 8;
			}

			throw new NotSupportedException(String.Format("Unknown bytes per component for \"{0}\".", format));
		}

		/// <summary>Returns the bytes per pixel for this pixel format.</summary>
		/// <param name="format">The pixel format.</param>
		public static int BytesPerPixel(this PixelFormat format)
		{
			return format.ComponentsPerPixel() * format.BytesPerComponent();
		}

		/// <summary>Returns the number of colour components per pixel.</summary>
		/// <param name="format">The pixel format.</param>
		public static int ComponentsPerPixel(this PixelFormat format)
		{
			PixelFormat colours = format & PixelFormat.LayoutMask;

			if(PixelFormat.None != (colours & PixelFormat.Grey        )) return 1;
			if(PixelFormat.None != (colours & PixelFormat.ThreeColours)) return 3;
			if(PixelFormat.None != (colours & PixelFormat.FourColours )) return 4;

			throw new NotSupportedException(String.Format("Unknown Format \"{0}\".", format));
		}

		/// <summary>Returns the datatype part of the given pixel format.</summary>
		/// <param name="format">The pixel format.</param>
		public static PixelFormat GetDataType(this PixelFormat format)
		{
			return format & PixelFormat.TypeMask;
		}

		/// <summary>Returns the layout part of the given pixel format.</summary>
		/// <param name="format">The pixel format.</param>
		public static PixelFormat GetLayout(this PixelFormat format)
		{
			return format & PixelFormat.LayoutMask;
		}

		/// <summary>Returns the maximum value of pixels of the given pixel format. Useful for converting between different pixel formats.</summary>
		/// <param name="format">The pixel format.</param>
		public static double MaxValue(this PixelFormat format)
		{
			var type = format & PixelFormat.TypeMask;
			switch (type)
			{
				case PixelFormat.Byte:   return Byte.MaxValue;
				case PixelFormat.Short:  return UInt16.MaxValue;
				case PixelFormat.Int:    return UInt32.MaxValue;
				case PixelFormat.Single: return 1.0;
				case PixelFormat.Double: return 1.0;
				default: throw new ArgumentException(
					String.Format("Type \"{0}\" is not supported (pixel format \"{1}\") not implemented.", type, format), "format");
			}
		}

		/// <summary>Returns the minimum value of pixels of the given pixel format. Useful for converting between different pixel formats.</summary>
		/// <param name="format">The pixel format.</param>
		public static double MinValue(this PixelFormat format)
		{
			var type = format & PixelFormat.TypeMask;
			switch (type)
			{
				case PixelFormat.Byte:   return 0.0;
				case PixelFormat.Short:  return 0.0;
				case PixelFormat.Int:    return 0.0;
				case PixelFormat.Single: return 0.0;
				case PixelFormat.Double: return 0.0;
				default: throw new ArgumentException(
					String.Format("Type \"{0}\" is not supported (pixel format \"{1}\") not implemented.", type, format), "format");
			}
		}
	}
}
