﻿// $Id: Pin.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;
using System.Diagnostics;
using System.Threading;

namespace VideoLib
{
	/// <summary>Possible states of a pin.</summary>
	public enum PinStatus
	{
		/// <summary>Unknown status. This is the default just after startup.</summary>
		Unknown,

		/// <summary>Ready to receive more data.</summary>
		Ready,

		/// <summary>Buffers are full.</summary>
		Blocked,
	}

	/// <summary>Event arguments encapsulating a change of pin status.</summary>
	[DebuggerDisplay("{OldStatus} => {NewStatus}")]
	public class PinStatusChangedEventArgs : EventArgs
	{
		/// <summary>Gets the old status.</summary>
		public PinStatus OldStatus { get; protected set; }

		/// <summary>Gets the new status.</summary>
		public PinStatus NewStatus { get; protected set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="PinStatusChangedEventArgs"/> class.
		/// </summary>
		/// <param name="oldStatus">The old status.</param>
		/// <param name="newStatus">The new status.</param>
		public PinStatusChangedEventArgs(PinStatus oldStatus, PinStatus newStatus)
		{
			OldStatus = oldStatus;
			NewStatus = newStatus;
		}
	}

	/// <summary>Delegate for handing status change events.</summary>
	public delegate void PinStatusChangedHandler(object sender, PinStatusChangedEventArgs e);

	/// <summary>Abstract base class of all IO pins.</summary>
	[DebuggerDisplay("{ToString()} ({Status})")]
	public abstract class Pin
	{
		/// <summary>Parent node of the IO Pin.</summary>
		public INode Node { get; protected set; }

		/// <summary>Name of this IO pin.</summary>
		public string Name { get; protected set; }

		/// <summary>Current status of this IO Pin.</summary>
		public virtual PinStatus Status
		{
			get { return status; }
			protected set
			{
				if (status != value)
				{
					PinStatus oldStatus = status;
					status = value;
					OnStatusChanged(oldStatus, status);
				}
			}
		}
		private PinStatus status = PinStatus.Unknown;

		/// <summary>Indicating whether this <see cref="Pin"/> produces verbose output.</summary>
		public bool Verbose { get; set; }

		/// <summary>Status changed event.</summary>
		public event PinStatusChangedHandler StatusChanged;

		/// <summary>New frame event.</summary>
		public event NewFrameEventHandler NewFrame;

		/// <summary>Constructs a new IO pin.</summary>
		/// <param name="name">Name of this IO pin.</param>
		/// <param name="node">Parent node of this IO pin.</param>
		public Pin(string name, INode node)
		{
			Name = name;
			Node = node;
			Verbose = false;
		}

		/// <summary>Invokes a PinStatusChanged event.</summary>
		/// <param name="oldStatus">Old status.</param>
		/// <param name="newStatus">New status.</param>
		protected virtual void OnStatusChanged(PinStatus oldStatus, PinStatus newStatus)
		{
			if(Verbose) Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, this, oldStatus, newStatus));
			if (StatusChanged != null)
				StatusChanged(this, new PinStatusChangedEventArgs(oldStatus, newStatus));
		}

		/// <summary>Invokes a NewFrame event.</summary>
		/// <param name="frame">The new frame.</param>
		protected void OnNewFrame(Frame frame)
		{
			if (NewFrame != null)
				NewFrame(this, new NewFrameEventArgs(frame));
		}

		/// <summary>Request the pin to update its status.</summary>
		public abstract void UpdateStatus();

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		public override string ToString()
		{
			return Node.ToString() + "." + this.Name;
		}

		/// <summary>Returns a hash code for this instance.</summary>
		public override int GetHashCode()
		{
			return Node.GetHashCode() ^ Name.GetHashCode();
		}
	}
}
