﻿// $Id: SimpleDequeScheduler.cs 293 2010-08-04 16:26:33Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using VideoLib.Frames;
using System.Diagnostics;
using VideoLib.Utils;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a deque-based scheduler.</summary>
	/// <remarks>
	/// This scheduler maintains a single double-ended queue of work items. Worker threads are
	/// constantly hungry for new work and run the front item of the queue. Nodes that become ready
	/// for processing are pushed at the back, to run them in-order. Frames to be transported are
	/// pushed at the front, to get them done asap.
	/// 
	/// Note that this may reorder frames, but it seems to work fine, as frames always have priority
	/// over nodes and hence are handled before the same node has the opportunity to produce the next
	/// frame.
	/// 
	/// IMPORTANT: This scheduler, as well as <see cref="CudaAwareScheduler"/> and
	/// <see cref="ModularScheduler"/> are the only working schedulers at the moment,
	/// due to changes in how events are handled.
	/// </remarks>
	public class SimpleDequeScheduler : Scheduler
	{
		/// <summary>Constructs a simple deque-based scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public SimpleDequeScheduler(Topology topology) : base(topology)
		{
		}

		/// <summary>Ready queue: queue of work items that are ready to run.</summary>
		private Deque<WorkItem> Ready = new Deque<WorkItem>();

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();

			try
			{
				// initialise status
				Status = SchedulerStatus.Starting;

				// listen to all new frames and status updates from output pins
				foreach (var edgeFan in Topology.OutgoingEdges)
				{
					edgeFan.Key.NewFrame += NewFrameListener;
					edgeFan.Key.StatusChanged += StatusChangedListener;
				}

				// listen to all status updates from input pins
				foreach (var edgeFan in Topology.IncomingEdges)
					edgeFan.Key.StatusChanged += StatusChangedListener;

				// listen to all status changes from nodes
				foreach (INode node in Topology.Nodes)
					node.StatusChanged += StatusChangedListener;

				// start up all nodes
				foreach (INode node in Topology.Nodes)
				{
					node.Init();
					node.UpdateStatus();
				}

				// let's go over to proper running mode
				Status = SchedulerStatus.Running;
			}
			catch (Exception)
			{
				throw;
			}
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID, if <see cref="Scheduler.StartWorker"/> is not not overidden.</param>
		protected override void RunWorker(object data)
		{
			// pin thread to processor cores
			SchedulerHelper.PinCurrentThreadToCore((int)data);

			// Run an infinite loop, looking for work to do
			while(Status == SchedulerStatus.Running)
			{
				// take first node from Ready queue and run it
				WorkItem workitem = null;
				lock (Ready)
				{
					// check if there is work to do (first chance)
					if (Ready.Count == 0) Monitor.Wait(Ready, 100);

					// check if there is work to do (second chance)
					if (Ready.Count == 0) continue;

					// get work item to run
					workitem = Ready.PopFront();
				}

				// run the work item
				workitem.Run();
			}

			SchedulerHelper.UnpinCurrentThread();
			Console.WriteLine(Thread.CurrentThread.Name + ": Stopped.");
		}

		/// <summary>Schedules transport of new frames to receivers.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		private void NewFrameListener(object sender, NewFrameEventArgs e)
		{
			OutputPin outputPin = (OutputPin)sender;
			foreach (var inputPin in Topology.OutgoingEdges[outputPin])
			{
				lock (Ready)
				{
					Ready.PushFront(new TransportWorkItem(outputPin, inputPin, e));
					Monitor.Pulse(Ready);
				}
			}
		}

		/// <summary>Listens to status changes in IO pins.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		private void StatusChangedListener(object sender, PinStatusChangedEventArgs e)
		{
			if (sender is InputPin) // then update upstream output pins
			{
				foreach (var outPin in Topology.IncomingEdges[(InputPin)sender])
				{
					lock (Ready)
					{
						Ready.PushFront(new ActionWorkItem(delegate() { outPin.DownStreamStatusChangedListener(sender, e); }));
						Monitor.Pulse(Ready);
					}
				}
				return;
			}

			if (sender is OutputPin) // then update node
			{
				OutputPin outputPin = (OutputPin)sender;

				lock (Ready)
				{
					Ready.PushFront(new ActionWorkItem(delegate() { outputPin.Node.UpdateStatus(); }));
					Monitor.Pulse(Ready);
				}
			}
		}

		private void StatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus));
			INode node = (INode)sender;

			// if a node is now ready to run, append it to the Ready queue
			if (e.NewStatus == NodeStatus.Ready)
			{
				lock (Ready)
				{
					Ready.PushBack(new ProcessingWorkItem(node) { Profiler = Profiler });
					Monitor.Pulse(Ready);
				}
				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting) return;

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running) return;

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished) return;

			// if a node is now stopped, exit everything
			if (e.NewStatus == NodeStatus.Stopped)
			{
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}
	}
}
