﻿// $Id: SharedWorkQueue.cs 440 2010-10-18 11:49:09Z cr333 $
// Partially based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using VideoLib.Frames;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a work queue that is shared between several worker threads.</summary>
	/// <remarks>Should be last in the list of work queues.</remarks>
	public class SharedWorkQueue : WorkQueue
	{
		/// <summary>The number of worker threads.</summary>
		private int numWorkers;

		/// <summary>Initializes a new instance of the <see cref="SharedWorkQueue"/> class.</summary>
		/// <param name="numWorkers">The number of workers threads to start.</param>
		public SharedWorkQueue(int numWorkers)
		{
			this.numWorkers = numWorkers;
		}

		/// <summary>Starts all worker threads associated with this work queue.</summary>
		/// <returns>Collection of all started worker threads.</returns>
		public override IEnumerable<Thread> StartWorkers()
		{
			for (int workerID = 0; workerID < numWorkers; workerID++)
				StartWorker(workerID, String.Format("Worker{0}", workerID));

			return workers;
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID.</param>
		protected override void RunWorker(object data)
		{
			// lock current managed thread to underlying OS thread (given by worker ID)
			SchedulerHelper.PinCurrentThreadToCore((int)data);

			base.RunWorker(data);

			SchedulerHelper.UnpinCurrentThread();
		}

		/// <summary>Handles initialisation of nodes.</summary>
		/// <param name="node">The node to be initialised.</param>
		/// <returns>True, if and only if the node has been initialised.</returns>
		public override bool HandleNodeInitEvent(INode node)
		{
			return HandleNodeInitEvent(node, Ready);
		}

		/// <summary>Handles new frame events.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public override bool HandleNewFrameEvent(object sender, NewFrameEventArgs e)
		{
			return HandleNewFrameEvent(sender, e, Topology, Ready);
		}

		/// <summary>Handles nodes that become ready to run.</summary>
		/// <param name="node">The ready node.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public override bool HandleNodeReadyEvent(INode node)
		{
			return HandleNodeReadyEvent(node, Scheduler.Profiler, Ready);
		}

		/// <summary>Handles changes of pin status.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public override bool HandlePinStatusChangedEvent(object sender, PinStatusChangedEventArgs e)
		{
			return HandlePinStatusChangedEvent(sender, e, Topology, Ready);
		}
	}
}