﻿// $Id: VideoPanel.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using VideoLib.Frames;

namespace VideoLib.Sinks
{
	/// <summary>Implements a double-buffered video display panel.</summary>
	public partial class VideoPanel : UserControl, IVideoControl
	{
		/// <summary>The back-buffer that contains the currently displayed content.</summary>
		private Bitmap backBuffer = null;

		/// <summary>Flag to store if any frames arrived so far.</summary>
		private bool gotImage = false;

		/// <summary>Instantiates a new double-buffered video display panel.</summary>
		public VideoPanel()
		{
			InitializeComponent();

			// set custom style
			this.SetStyle(ControlStyles.DoubleBuffer, true);
			this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
			this.SetStyle(ControlStyles.UserPaint, true);
			this.SetStyle(ControlStyles.Selectable, false);
			this.SetStyle(ControlStyles.ResizeRedraw, true);

			// initialise the back-buffer with the current control size
			VideoSize = new Size(Width, Height);
			backBuffer = new Bitmap(Width, Height);
		}
		
		/// <summary>Indicates if the control is currently displayed in the designer.</summary>
		protected bool InDesignMode
		{
			get { if (DesignMode) return true; return this.ParentForm.Site != null; }
		}

		private Size videoSize;
		/// <summary>Gets or sets the size of the video.</summary>
		public Size VideoSize
		{
			get { return videoSize; }
			set
			{
				if (videoSize != value)
				{
					videoSize = value;
					if (!Disposing && Created)
						BeginInvoke(new Action(delegate() {
							this.Size = videoSize;  MinimumSize = videoSize; MaximumSize = videoSize;
						}));
				}
			}
		}

		/// <summary>Accepts a new frame and caption to be drawn onto this control.</summary>
		/// <param name="frame">The new frame.</param>
		/// <param name="caption">The new caption.</param>
		public void Draw(Frame frame, string caption)
		{
			Draw(((BitmapFrame)frame).Bitmap, caption);
		}

		/// <summary>Accepts a new bitmap and caption to be drawn onto this control.</summary>
		/// <param name="bitmap">The new bitmap.</param>
		/// <param name="caption">The new caption.</param>
		public void Draw(Bitmap bitmap, string caption)
		{
			gotImage = true;
			lock (backBuffer) // to avoid issues when OnPaint runs in another thread
			{
				// allocate new back-buffer if size changes
				if (backBuffer.Size != bitmap.Size)
				{
					VideoSize = bitmap.Size;
					backBuffer = new Bitmap(bitmap.Width, bitmap.Height);
				}

				// draw to back-buffer
				using (Graphics g = Graphics.FromImage(backBuffer))
				{
					g.Clear(BackColor);
					g.DrawImage(bitmap, 0, 0);
					g.DrawString(caption, Font, new SolidBrush(Color.LightGreen), 0, 0);
				}
			}

			// asynchronously change size of & redraw the control
			if (!Disposing && Created)
				BeginInvoke(new Action(delegate() { Refresh(); }));
		}

		/// <summary>
		/// Overrided - ignore background painting, to avoid flickering
		/// </summary>
		/// <param name="e"></param>
		protected override void OnPaintBackground(PaintEventArgs e)
		{
			// ignore background painting, to avoid flickering
		}

		/// <summary>Custom handling of Paint events, to support Designer mode.</summary>
		/// <param name="e">Paint event arguments.</param>
		protected override void OnPaint(PaintEventArgs e)
		{
			// custom film strip look for Designer
			if (InDesignMode)
			{
				e.Graphics.Clear(BackColor);
				int framePadding = 4;
				int holePadding = 6;
				int holeSize = 10;

				// black background
				e.Graphics.FillRectangle(new SolidBrush(Color.Black), ClientRectangle);

				// punched holes
				for (int y = 4; y < Height; y += 2 * holeSize)
				{
					e.Graphics.FillRectangle(new SolidBrush(BackColor), holePadding, y, holeSize, holeSize);
					e.Graphics.FillRectangle(new SolidBrush(BackColor), Width - holePadding - holeSize, y, holeSize, holeSize);
				}

				// frames
				int frameBoundary1 = (int)(0.1 * Height + 0.5);
				int frameBoundary2 = (int)(0.9 * Height + 0.5);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, 0, Width - 4 * holePadding - 2 * holeSize, frameBoundary1 - framePadding / 2);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary1 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, frameBoundary2 - frameBoundary1 - framePadding);
				e.Graphics.FillRectangle(new SolidBrush(Color.LightGray), 2 * holePadding + holeSize, frameBoundary2 + framePadding / 2, Width - 4 * holePadding - 2 * holeSize, Height - (frameBoundary2 + framePadding / 2));

				// "Video Panel" text
				Font font = new Font(Font.FontFamily, 16, FontStyle.Bold);
				String text = "Video" + Environment.NewLine + "Panel";
				StringFormat stringFormat = new StringFormat { Alignment = StringAlignment.Center, LineAlignment = StringAlignment.Center };
				e.Graphics.DrawString(text, font, new SolidBrush(ForeColor), Width / 2, Height / 2, stringFormat);
			}
			else if (!gotImage) // what to show while waiting for images to display
			{
				// white background
				e.Graphics.FillRectangle(new SolidBrush(Color.White), ClientRectangle);

				// outline
				e.Graphics.DrawRectangle(new Pen(Color.Silver), 0, 0, ClientSize.Width - 1, ClientSize.Height - 1);

				// "waiting ..." text
				Font font = new Font(Font.FontFamily, Font.Size, FontStyle.Italic);
				String text = "waiting for video …";
				StringFormat stringFormat = new StringFormat { Alignment = StringAlignment.Center, LineAlignment = StringAlignment.Center };
				e.Graphics.DrawString(text, font, new SolidBrush(Color.DarkGray), Width / 2, Height / 2, stringFormat);
			}

			if (backBuffer != null)
			{
				lock (backBuffer) // to avoid issues with DrawImage running on a different thread
					e.Graphics.DrawImageUnscaled(backBuffer, 0, 0);
			}
		}
	}
}
